import os

#from envpath.settings import *

help_function_dir, this_filename = os.path.split(__file__)
default_feature_dir='.' #The dictory for .bed files, which can be modified as needed
source_script_dir = '.' #The dictory for Rscripts, which can be modified as needed

color_list = [
    "#1f78b4",
    "#ff7f00",
    "#33a02c",
    "#e31a1c",
    "#6a3d9a",
    "#8c564b",
    "#e377c2",
    "#969696",  #
    "#17becf",
    "#253494",
    #'#ffffcc',
    "#a6cee3",
    "#b2df8a",
    "#fb9a99",
    # '#fdbf6f',
    "#cab2d6",
    #'#ffff99',
    "#99d8c9",
    "#8c96c6",
    "#66c2a5",
    "#636363",
    "#ffd92f",
    "#b3b3b3",
    "#db5f57",
]


color_list_2 = [
    "#8dd3c7",
    "#bebada",
    "#fb8072",
    "#80b1d3",
    "#fdb462",
    "#b3de69",
    "#fccde5",
    "#d9d9d9",
    "#bc80bd",
    "#ccebc5",
    "#ffed6f",
    "#ffffb3",
    "#99d8c9",
    "#8c96c6",
    "#66c2a5",
    "#636363",
    "#ffd92f",
    "#b3b3b3",
    "#db5f57",
]



import os

import numpy as np
import pandas as pd
import pyranges as pr
from matplotlib import pyplot as plt

from .settings import *


def generate_singleCpG_within_region_annotation(
    selected_region_annotation="500bp", CpG_ref="mm10", recompute_annotation=False
):
    singleCpG_anno_path = (
        f"{default_feature_dir}/{selected_region_annotation}_singleCpG.bed"
    )
    print(f"singleCpG_anno_path: {singleCpG_anno_path}")

    if not os.path.exists(singleCpG_anno_path) or recompute_annotation:
        print(f"generate new {selected_region_annotation}_singleCpG.bed file")

        if CpG_ref not in ["hg38", "hg19", "mm10", "mm9"]:
            raise ValueError(f"CpG_ref {CpG_ref} is not found")
        CpG_ref_dir = f"{CpG_ref_location}/{CpG_ref}.CpGs_new.bed"

        df_ref = pd.read_csv(
            CpG_ref_dir,
            sep="\t",
            names=["Chromosome", "Start", "name", "chr_pos", "chr_pos_neg"],
        )

        df_ref["End"] = df_ref["Start"] + 1

        df_region = pd.read_csv(
            f"{default_feature_dir}/{selected_region_annotation}.bed",
            sep="\t",
            header=None,
        )
        df_region.columns = [
            "Chromosome",
            "Start",
            "End",
            "sign",
            "chr_start_end",
            "length",
        ]

        chr_dict = {str(x): f"chr{x}" for x in range(1, 30)}
        chr_dict.update({f"chr{x}": f"chr{x}" for x in range(1, 30)})
        chr_dict.update({"X": "chrX", "Y": "chrY", "chrX": "chrX", "chrY": "chrY"})

        df_region["Chromosome"] = df_region["Chromosome"].map(chr_dict)
        df_region = df_region[~pd.isna(df_region["Chromosome"])]

        pr_df_ref = pr.PyRanges(df_ref)
        pr_df_region = pr.PyRanges(df_region)
        df_intersect = pr_df_ref.join(pr_df_region).as_df()
        df_intersect = df_intersect.filter(["chr_pos", "chr_pos_neg"]).drop_duplicates()

        all_chr_pos = (
            df_intersect["chr_pos"].tolist() + df_intersect["chr_pos_neg"].tolist()
        )
        df_all_chr_pos = pd.DataFrame(all_chr_pos, columns=["chr_pos"])
        df_all_chr_pos[["Chromosome", "Start"]] = df_all_chr_pos["chr_pos"].str.split(
            "_", expand=True
        )
        del df_all_chr_pos["chr_pos"]
        df_all_chr_pos["Start"] = df_all_chr_pos["Start"].astype(int)
        df_all_chr_pos["End"] = df_all_chr_pos["Start"] + 1

        df_all_chr_pos["sign"] = "*"
        df_all_chr_pos["chr_start_end"] = (
            "singleCpG_"
            + df_all_chr_pos["Chromosome"].astype(str)
            + "_"
            + df_all_chr_pos["Start"].astype(str)
        )
        df_all_chr_pos["label"] = "singleCpG_region_selection"

        df_all_chr_pos.to_csv(singleCpG_anno_path, sep="\t", header=None, index=False)

    df_CpG = pd.read_csv(
        singleCpG_anno_path,
        sep="\t",
        header=None,
    )
    df_CpG.columns = [
        "Chromosome",
        "Start",
        "End",
        "sign",
        "chr_start_end",
        "label",
    ]
    df_CpG = df_CpG[~pd.isna(df_CpG["Chromosome"])]

    selected_CpG = [
        f"{x}_{y}"
        for x, y in zip(df_CpG["Chromosome"].to_list(), df_CpG["Start"].to_list())
    ]

    return selected_CpG


def zscore(df, key="rate"):
    mean = df[key].mean()
    std = df[key].std()
    return (df[key] - mean) / std


def find_signal_region(
    df, n=5, plot=False, win_thresh=1, consecutive_N=3, peak_height=1.5
):
    import pyranges as pr
    from scipy.signal import argrelextrema

    # n = 5,   number of points to be checked before and after
    ## Find local peaks
    df["max"] = df.iloc[
        argrelextrema(df.rate_zscore.values, np.greater_equal, order=n)[0]
    ]["rate_zscore"]

    df.loc[(df["max"] < peak_height), "max"] = np.nan

    if plot:
        # Plot results
        fig, ax = plt.subplots(figsize=(15, 4))
        df_plot = df.iloc[600:700]
        # plt.scatter(df_plot.index, df_plot['min'], c='r')
        plt.plot(df_plot.index, df_plot["rate_zscore"], "-*r")
        plt.scatter(df_plot.index, df_plot["max"], c="g", s=100)
        plt.show()

    ## Find consecutive positive region
    df_diff = (df["rate_zscore"] > win_thresh).astype(int).diff()
    df_start = df[df_diff == 1]
    df_end = df[df_diff == -1]
    win_size = df.iloc[0]["end"] - df.iloc[0]["start"] + 1
    df_start["end"] = df_start["start"].shift(periods=-1).fillna(10**9).astype(int)
    df_end["end"] = df_end["start"]
    pr_start = pr.PyRanges(
        df_start.rename(columns={"chr": "Chromosome", "start": "Start", "end": "End"})
    )
    pr_end = pr.PyRanges(
        df_end.rename(columns={"chr": "Chromosome", "start": "Start", "end": "End"})
    )
    df_bed = (
        pr_start.join(pr_end)
        .as_df()
        .filter(["Chromosome", "Start", "End_b"])
        .rename(columns={"End_b": "End"})
    )
    df_bed["len"] = df_bed["End"] - df_bed["Start"]
    df_bed["id"] = (
        df_bed["Chromosome"].astype(str)
        + "_"
        + df_bed["Start"].astype(str)
        + "_"
        + df_bed["End"].astype(str)
    )
    df_bed = df_bed[df_bed["len"] / win_size >= consecutive_N]

    # interset with high peaks
    df_pr_1 = pr.PyRanges(
        df[df["max"] > peak_height].rename(
            columns={"chr": "Chromosome", "start": "Start", "end": "End"}
        )
    )
    df_pr_2 = pr.PyRanges(df_bed)
    df_out = df_pr_2.join(df_pr_1).as_df()

    df_out = df_out.set_index("id")
    df_out["max_rate"] = df_out.groupby(["id"]).agg(max_rate=("rate_zscore", "max"))

    df_out = (
        df_out.filter(["Chromosome", "Start", "End", "len", "max_rate"])
        .drop_duplicates()
        .reset_index()[["Chromosome", "Start", "End", "len", "max_rate", "id"]]
    )
    return df_out


#########

# peaks

#########


def find_signal_peaks_v0(
    df, n=5, plot=False, win_thresh=0.5, consecutive_N=3, peak_height=1.5
):
    import pyranges as pr
    from scipy.signal import argrelextrema

    # n = 5,   number of points to be checked before and after
    ## Find local peaks
    df["max"] = df.iloc[
        argrelextrema(df.rate_zscore.values, np.greater_equal, order=n)[0]
    ]["rate_zscore"]

    df.loc[(df["max"] < peak_height), "max"] = np.nan

    if plot:
        # Plot results
        fig, ax = plt.subplots(figsize=(15, 4))
        df_plot = df.iloc[600:700]
        # plt.scatter(df_plot.index, df_plot['min'], c='r')
        plt.plot(df_plot["start"], df_plot["rate_zscore"], "-*r")
        plt.scatter(df_plot["start"], df_plot["max"], c="g", s=100)
        plt.show()

    ## Find consecutive positive region
    df_diff = (df["rate_zscore"] > win_thresh).astype(int).diff()
    df_start = df[df_diff == 1]
    df_end = df[df_diff == -1]
    win_size = df.iloc[0]["end"] - df.iloc[0]["start"] + 1
    df_start["end"] = df_start["start"].shift(periods=-1).fillna(10**9).astype(int)
    df_end["end"] = df_end["start"]
    pr_start = pr.PyRanges(
        df_start.rename(columns={"chr": "Chromosome", "start": "Start", "end": "End"})
    )
    pr_end = pr.PyRanges(
        df_end.rename(columns={"chr": "Chromosome", "start": "Start", "end": "End"})
    )
    df_bed = (
        pr_start.join(pr_end)
        .as_df()
        .filter(["Chromosome", "Start", "End_b"])
        .rename(columns={"End_b": "End"})
    )
    df_bed["len"] = df_bed["End"] - df_bed["Start"]
    df_bed["id"] = (
        df_bed["Chromosome"].astype(str)
        + "_"
        + df_bed["Start"].astype(str)
        + "_"
        + df_bed["End"].astype(str)
    )
    df_bed = df_bed[df_bed["len"] / win_size >= consecutive_N]

    # interset with high peaks
    df_pr_1 = pr.PyRanges(
        df[df["max"] > peak_height].rename(
            columns={"chr": "Chromosome", "start": "Start", "end": "End"}
        )
    )
    df_pr_2 = pr.PyRanges(df_bed)
    df_out = df_pr_1.join(df_pr_2).as_df()

    #     df_out=df_out.set_index('id')
    #     df_out['max_rate']=df_out.groupby(['id']).agg(max_rate=('rate_zscore','max'))

    #     df_out=df_out.filter(['Chromosome','Start','End','len','max_rate']
    #              ).drop_duplicates().reset_index()[['Chromosome','Start','End','len','max_rate','id']]
    return df_out


def find_signal_peaks(
    df_data, source="acc", tile_window_size=100, smooth_size=5, zscore_thresh=2.5
):
    """
    We first tile the genome with tile_window_size, and average the rate within the window, then we locally smooth the
    rate using neighboring smoooth_size locus. Then, we z-score transform the smoothed rate, and using the cutoff zscore_thresh to select the signal region. The intermediate results are inserted to the output df_tile

    source: {'acc', 'met'}
    """
    if "Start" not in df_data.columns:
        df_data["Start"] = df_data["pos"].astype(int)
        df_data["End"] = df_data["pos"].astype(int)
    if "Chromosome" not in df_data.columns:
        df_data["Chromosome"] = df_data["chr"]

    # define the available size of each chromosome
    print(f"Generate {tile_window_size}bp genome tiles and perform intersection")
    df_backbone = df_data.groupby("Chromosome").agg(End=("End", "max")).reset_index()
    df_backbone["Start"] = 0
    df_backbone = df_backbone[df_backbone["Chromosome"] != "chrlambda"]

    df_tile = (
        pr.gf.tile_genome(pr.PyRanges(df_backbone), tile_window_size, tile_last=False)
        .join(pr.PyRanges(df_data))
        .as_df()
    )

    # change the type from catogories to str, to avoid expanding the operation
    # to non-existing chromosomes (which are still trakced in df.cat.categories)

    df_list = []
    for chr_tmp in sorted(df_tile["Chromosome"].cat.categories):
        print(chr_tmp)
        df_tmp = df_tile[df_tile["Chromosome"] == chr_tmp]
        if source == "acc":
            df_tmp = (
                df_tmp.groupby(["Start"])
                .agg(tot_met=("met_sites", "sum"), tot_nonmet=("nonmet_sites", "sum"))
                .assign(rate=lambda x: x["tot_met"] / (x["tot_met"] + x["tot_nonmet"]))
                .reset_index()
            )
        else:
            df_tmp = (
                df_tmp.groupby(["Start"])
                .agg(tot_met=("met_sites", "sum"), tot_nonmet=("nonmet_sites", "sum"))
                .assign(
                    rate=lambda x: 1 - x["tot_met"] / (x["tot_met"] + x["tot_nonmet"])
                )
                .reset_index()
            )
        df_tmp["tot_met"] = df_tmp["tot_met"].astype(int)
        df_tmp["tot_N"] = df_tmp["tot_met"] + df_tmp["tot_nonmet"]

        df_tmp["smooth_rate"] = df_tmp.rolling(
            window=smooth_size, center=True, min_periods=1
        ).agg({"rate": "mean"})

        df_tmp["smooth_rate_zscore"] = zscore(df_tmp, key="smooth_rate")
        df_tmp["signal"] = df_tmp["smooth_rate_zscore"]
        df_tmp.loc[df_tmp["smooth_rate_zscore"] < zscore_thresh, "signal"] = 0
        df_tmp["End"] = df_tmp["Start"] + tile_window_size
        df_tmp["Chromosome"] = chr_tmp

        df_list.append(df_tmp)
    df_out = pd.concat(df_list)
    return df_out


def generate_feature_bed_file_from_signal_peaks(
    df_tile,
    save_name=None,
    signal_key="signal",
    merge=True,
    annotation_name="peak",
):
    """
    Take output from find_signal_peaks, and merge neighboring peaks,
    and convert to a feature bed file saved at given name
    """
    if merge:
        df_bed = pr.PyRanges(df_tile).merge().as_df()
    else:
        df_bed = pr.PyRanges(df_tile).as_df()

    df_bed["Strand"] = "*"
    df_bed["id"] = (
        df_bed["Chromosome"].astype(str)
        + "_"
        + df_bed["Start"].astype(str)
        + "_"
        + df_bed["End"].astype(str)
    )
    df_bed["Annotation"] = annotation_name
    df_bed = df_bed.filter(["Chromosome", "Start", "End", "Strand", "id", "Annotation"])

    if save_name != None:
        print(f"file saved at:\n", f"{default_feature_dir}/{save_name}.bed")
        df_bed.to_csv(
            f"{default_feature_dir}/{save_name}.bed", sep="\t", index=0, header=None
        )

    return df_bed


def tile_stats(df_data, tile_window_size=5000):
    """
    Aggregate data within the tile bin of size tile_window_size,
    and check how many peaks and cell types within each bin

    df_data should have columns ['Chromosome','Start','cell_type']
    """
    assert np.isin(["Chromosome", "Start", "End", "cell_type"], df_data.columns).all()

    if "id" not in df_data.columns:
        df_data["id"] = (
            df_data["Chromosome"].astype(str)
            + "_"
            + df_data["Start"].astype(str)
            + "_"
            + df_data["End"].astype(str)
        )

    df_tile_2 = (
        pr.gf.tile_genome(pr.data.chromsizes(), tile_window_size, tile_last=False)
        .join(pr.PyRanges(df_data))
        .as_df()
    )

    df_new_2 = (
        df_tile_2.groupby(["Chromosome", "Start"])
        .agg(
            celltype_N=("cell_type", lambda x: len(set(x))),
            celltypes=("cell_type", lambda x: ",".join(set(x))),
            peak_N=("id", lambda x: len(set(x))),
        )
        .reset_index()
    )
    df_new_2["End"] = df_new_2["Start"] + tile_window_size
    df_new_2 = df_new_2.dropna()
    return df_new_2


def remove_overlapping(df):
    sel_idx = df.iloc[1:]["start"].to_numpy() >= df.iloc[:-1]["end"].to_numpy()
    sel_idx = [True] + list(sel_idx)
    return df[sel_idx]  # .filter(['start','end','value'])



from .adata import *
from .analysis import *
from .batch import *
from .clone import *
from .lineage import *
from .metadata import *
from .plotting import *
from .QC import *
from .region import *
from .scripts import *
from .settings import *
from .similarity import *



import glob
import os

import numpy as np
import pandas as pd
import scipy.sparse as ssp
import yaml

from .settings import *

## sample operations, and loading various annotation and data file, transfer labels etc.

######################

# change or update with df_sample

######################


def update_sample_info_on_adata(adata, df_sample):
    df_sample = df_sample.set_index("sample")
    shared_frac = (
        len(set(df_sample.index).intersection(adata.obs_names)) / adata.shape[0]
    )
    if shared_frac < 0.5:
        print("adata and df_sample have <50 percent shared values")
    for x in df_sample.columns:
        adata.obs[x] = df_sample[x]
    df_sample = df_sample.reset_index()


def backup_and_save_sample_info(df_sample, root_dir):
    # get the current versions for saving
    file_path = f"{root_dir}/sample_sheet.tsv.gz"
    os.system(f"mkdir -p {root_dir}/old_sample_sheet")
    all_paths = sorted(glob.glob(f"{root_dir}/old_sample_sheet/sample_sheet*.tsv.gz"))
    versions = [
        x.split("/")[-1].split("sample_sheet")[1].split(".tsv.gz")[0] for x in all_paths
    ]
    old_version = [x.split("_v")[1] for x in versions if "_v" in x]
    if len(old_version) == 0:
        cur_version = 1
    else:
        cur_version = np.max(np.array(old_version).astype(int)) + 1

    if os.path.exists(file_path):
        # rename the current version
        os.system(
            f"mv {file_path} {root_dir}/old_sample_sheet/sample_sheet_v{cur_version}.tsv.gz"
        )
        # print(f"mv {file_path} {root_dir}/old_sample_sheet/sample_sheet_v{cur_version}.tsv.gz")
    # save a new version
    df_sample.to_csv(file_path, sep="\t", index=False, compression="gzip")


def initialize_sample_info(root_dir):
    """
    We do not save sample info here
    """
    print("Initialize sample info")
    with open(f"{root_dir}/../../config.yaml", "r") as stream:
        file = yaml.safe_load(stream)
        SampleList = file["SampleList"]
    true_list = [True for __ in range(len(SampleList))]
    lineage = ["1" for __ in range(len(SampleList))]
    df_sample = pd.DataFrame(
        {
            "sample": SampleList,
            "cell": SampleList,
            "pass_accQC": true_list,
            "pass_metQC": true_list,
            "id_met": SampleList,
            "id_acc": SampleList,
            "lineage": lineage,
            "stage": lineage,
        }
    )
    return df_sample


def update_samples_from_config(root_dir):
    df_ini_sample_info = initialize_sample_info(root_dir)
    df_sample = load_sample_info(root_dir, force_run=False)
    df_sample_merge = pd.concat(
        [df_sample, df_ini_sample_info], ignore_index=True
    ).drop_duplicates("sample")
    backup_and_save_sample_info(df_sample_merge, root_dir)
    print(
        f"old_samples: {len(df_sample)}; config samples: {len(df_ini_sample_info)}; merged samples: {len(df_sample_merge)}"
    )
    return df_sample_merge


def update_sample_info_with_RNA(root_dir, adata_rna, id_rna_map=None, save=False):
    """
    The root directory should be leading to a DNA folder with 'met, acc'

    id_rna_map: a dictionary to map 'sample' to 'id_rna'
    """

    # df_sample = pd.read_csv(
    #     f"{root_dir}/results/metacc/qc/sample_metadata_after_metacc_qc.txt.gz",
    #     sep="\t",
    #     compression="gzip",
    # )
    df_sample = load_sample_info(root_dir)

    if id_rna_map is None:
        df_sample["plate"] = df_sample["sample"].apply(lambda x: x.split("_DNA_")[0])
        df_sample["plate_barcode"] = df_sample["sample"].apply(
            lambda x: x.split("_DNA_")[1]
        )

        Lime_barcode = mapping_from_plate_barcode_to_Lime_barcode(
            df_sample["plate_barcode"]
        )

        df_sample["Lime_barcode"] = Lime_barcode
        df_sample["id_rna"] = df_sample["plate"] + "_RNA_" + df_sample["Lime_barcode"]
    else:
        df_sample["id_rna"] = df_sample["sample"].map(id_rna_map)

    df_sample["pass_rnaQC"] = df_sample["id_rna"].isin(adata_rna.obs_names)
    df_sample["HQ"] = df_sample["pass_accQC"] & df_sample["pass_metQC"]

    ## Add RNA annotations
    adata_rna_tmp = adata_rna[adata_rna.obs_names.isin(df_sample["id_rna"])]
    df_sample = df_sample.set_index("id_rna")
    df_UMAP = pd.DataFrame(
        {
            "id_rna": adata_rna_tmp.obs_names,
            "UMAP_rna_x": adata_rna_tmp.obsm["X_umap"][:, 0],
            "UMAP_rna_y": adata_rna_tmp.obsm["X_umap"][:, 1],
        }
    ).set_index("id_rna")
    df_sample.loc[adata_rna_tmp.obs_names, "celltype"] = adata_rna_tmp.obs[
        "cell_type"
    ].astype(str)
    df_sample.loc[adata_rna_tmp.obs_names, "leiden"] = adata_rna_tmp.obs[
        "leiden"
    ].astype(str)
    df_sample.loc[adata_rna_tmp.obs_names, "UMAP_rna_x"] = df_UMAP["UMAP_rna_x"]
    df_sample.loc[adata_rna_tmp.obs_names, "UMAP_rna_y"] = df_UMAP["UMAP_rna_y"]

    df_sample.loc[pd.isna(df_sample["celltype"]), "celltype"] = "nan0"
    df_sample.loc[df_sample["celltype"] == "nan", "celltype"] = "nan0"
    df_sample.loc[pd.isna(df_sample["leiden"]), "leiden"] = "nan0"
    df_sample.loc[df_sample["leiden"] == "nan", "leiden"] = "nan0"

    df_sample["lineage"] = df_sample["celltype"]
    df_sample["lineage"] = [
        "_".join(x.split(" ")) for x in df_sample["lineage"]
    ]  # remove any spaces
    df_sample["stage"] = df_sample["lineage"]
    df_sample["cell"] = df_sample["lineage"]

    df_sample = df_sample.reset_index()
    if save:
        print("save the sample_sheet and rna")
        backup_and_save_sample_info(df_sample, root_dir)
        os.makedirs(f"{root_dir}/rna", exist_ok=True)
        adata_rna.write_h5ad(f"{root_dir}/rna/rna_adata.h5ad")
    return df_sample


def get_plate_barcode_to_Lime_barcode_map():
    df_barcode_map = pd.read_csv(
        f"{help_function_dir}/barcode_map.csv"
    ).filter(["LimeCat_BC", "BC_id_num"])
    num_to_BC = dict(zip(df_barcode_map["BC_id_num"], df_barcode_map["LimeCat_BC"]))

    count = 0
    mapping = {}  # map from a plate ID to the Lime barcode.
    for x in range(1, 13):
        for y in ["A", "B", "C", "D", "E", "F", "G", "H"]:
            count = count + 1
            mapping[f"{y}{x}"] = num_to_BC[
                count
            ]  # map from a plate ID like A10 to a barcode ordering number, and then to the Lime barcode.

    return mapping


def mapping_from_Lime_barcode_to_plate_ID(Lime_barcode):
    XX = get_plate_barcode_to_Lime_barcode_map()
    mapping = dict(zip(XX.values(), XX.keys()))
    return [mapping[x] for x in Lime_barcode]


def mapping_from_plate_barcode_to_Lime_barcode(plate_barcode):
    mapping = get_plate_barcode_to_Lime_barcode_map()
    return [mapping[x] for x in plate_barcode]


def get_methscan_sample_index_map(compact_file_dir):
    df_stats = pd.read_csv(f"{compact_file_dir}/cell_stats.csv")
    df_stats.columns = ["sample", "nCG", "n_meth", "met_rate"]
    df_stats["index"] = np.arange(len(df_stats)) + 1
    sample_index_map = dict(zip(df_stats["sample"], df_stats["index"]))
    return sample_index_map


###########

# load data

###########


def load_sample_info(root_dir, force_run=False):
    """
    Load or initialize (force_run=True) sample_info.
    Note that this is the only place where we use the config.yaml file to
    define the SampleList. The rest of functions will only call load_sample_info
    to define SampleList

    Parameters
    ----------
        root_dir:
            Directory at the downstream_R/all_data/. By default, root_dir will point
            to this place throughput this file
    """
    file_path = f"{root_dir}/sample_sheet.tsv.gz"
    if os.path.exists(file_path) and (not force_run):
        df_sample = pd.read_csv(file_path, sep="\t", compression="gzip")
    else:
        df_sample = initialize_sample_info(root_dir)
        backup_and_save_sample_info(df_sample, root_dir)
    return df_sample


def load_barcode_map():
    df_barcode_map = pd.read_csv(
        f"{help_function_dir}/barcode_map.csv", index_col=0
    )
    return df_barcode_map


def load_gene_map():
    gene_map = pd.read_csv(
        f"{default_feature_dir}/../genes/Mmusculus_genes_BioMart.101.txt", sep="\t"
    )
    gene_map.rename(columns={"ens_id": "id"}, inplace=True)
    return gene_map


def get_gene_id_to_name_dict():
    df_map = load_gene_map()
    return dict(zip(df_map["id"], df_map["symbol"]))


def load_annotation(anno):
    data_dir = f"{default_feature_dir}/{anno}.bed"
    return pd.read_csv(
        data_dir,
        sep="\t",
        header=None,
        names=["chr", "start", "end", "strand", "ens_id", "feature"],
    )


def load_met(root_dir, anno):
    df_met = pd.read_csv(
        f"{root_dir}/met/feature_level/{anno}.tsv.gz",
        sep="\t",
        compression="gzip",
        header=None,
    )
    df_met = df_met.rename(
        columns={0: "sample", 1: "id", 2: "anno", 3: "Nmet", 4: "N", 5: "rate"}
    )
    df_met["Nmet"] = df_met["Nmet"].astype(int)
    df_met["N"] = df_met["N"].astype(int)
    df_met["rate"] = df_met["rate"].astype(float)
    if len(df_met.columns) > 6:
        df_met = df_met.rename(columns={6: "rate_before_correction"})
        df_met["rate_before_correction"] = df_met["rate_before_correction"].astype(
            float
        )
    return df_met


def load_acc(root_dir, anno):
    df_acc = pd.read_csv(
        f"{root_dir}/acc/feature_level/{anno}.tsv.gz",
        sep="\t",
        compression="gzip",
        header=None,
    )
    df_acc = df_acc.rename(
        columns={0: "sample", 1: "id", 2: "anno", 3: "Nmet", 4: "N", 5: "rate"}
    )
    df_acc["Nmet"] = df_acc["Nmet"].astype(int)
    df_acc["N"] = df_acc["N"].astype(int)
    df_acc["rate"] = df_acc["rate"].astype(float)
    if len(df_acc.columns) > 6:
        df_acc = df_acc.rename(columns={6: "rate_before_correction"})
        df_acc["rate_before_correction"] = df_acc["rate_before_correction"].astype(
            float
        )

    return df_acc


def load_rna(root_dir):
    import scanpy as sc

    adata = sc.read(f"{root_dir}/rna/rna_adata.h5ad")
    return adata


def merge_metaccrna(root_dir, anno, gene_map=None):
    # load meta data
    sample_info = pd.read_csv(f"{root_dir}/sample_sheet.tsv.gz", sep="\t")

    # load data
    df_met = load_met(root_dir, anno).filter(
        ["sample", "id", "anno", "Nmet", "N", "rate"]
    )
    df_met.rename(
        columns={"Nmet": "met_Nmet", "N": "met_N", "rate": "met_rate"}, inplace=True
    )

    df_acc = load_acc(root_dir, anno).filter(
        ["sample", "id", "anno", "Nmet", "N", "rate"]
    )
    df_acc.rename(
        columns={"Nmet": "acc_Nmet", "N": "acc_N", "rate": "acc_rate"}, inplace=True
    )

    df_metacc = (
        df_acc.merge(df_met.drop("anno", axis=1), on=["sample", "id"])
        .merge(
            sample_info.filter(
                [
                    "celltype",
                    "stage",
                    "sample",
                    "id_rna",
                    "pass_metQC",
                    "pass_rnaQC",
                    "pass_accQC",
                    "HQ",
                ]
            ),
            on="sample",
        )
        .query("HQ==True")
    )
    if gene_map is None:
        df_gene = load_gene_map()
        gene_map = dict(zip(df_gene["id"], df_gene["symbol"]))
        df_metacc["symbol"] = df_metacc["id"].map(gene_map)
    else:
        df_metacc["Unique_ID"] = anno.split("_nb")[0] + "@" + df_metacc["id"]
        df_metacc["symbol"] = df_metacc["Unique_ID"].map(gene_map)

    adata_rna = load_rna(root_dir)
    df_rna = adata_rna.to_df()

    df_rna = (
        df_rna.reset_index()
        .melt(id_vars=["index"], ignore_index=False)
        .rename(columns={"index": "id_rna", "variable": "symbol", "value": "rna_exp"})
    )

    df_metaccrna = df_metacc.merge(df_rna, on=["id_rna", "symbol"])  # , how="outer")
    return df_metaccrna


######################

# Additional annotation functions

######################


def match_index_between_adata(adata_source, adata_target):
    df_source = pd.DataFrame(
        {
            "id_rna": adata_source.obs_names,
            "source_idx": np.arange(adata_source.shape[0]).astype(int),
        }
    )
    df_target = pd.DataFrame(
        {
            "id_rna": list(adata_target.obs_names),
            "target_idx": np.arange(adata_target.shape[0]).astype(int),
        }
    )
    df_merge = df_source.merge(df_target, on="id_rna")

    # extract a common set of cells
    sel_idx = np.array(df_merge["source_idx"]).astype(int)
    adata_source_v1 = adata_source[sel_idx]
    sel_idx = np.array(df_merge["target_idx"]).astype(int)
    adata_target_v1 = adata_target[sel_idx]
    return adata_source_v1, adata_target_v1


def transfer_labels_from_rna_adata_to_DNA_adata(adata_acc, adata_rna, root_dir):
    import scanpy as sc

    if not ssp.issparse(adata_acc.X):
        adata_acc.X = ssp.csr_matrix(adata_acc.X)

    df_acc = pd.DataFrame(
        {
            "sample": adata_acc.obs_names,
            "acc_idx": np.arange(adata_acc.shape[0]).astype(int),
        }
    )
    df_sample = load_sample_info(root_dir)
    df_acc = df_acc.merge(df_sample, on="sample")
    adata_acc_v0 = adata_acc.copy()
    adata_acc_v0.obs_names = np.array(df_acc["id_rna"].astype(str))
    df_RNA = pd.DataFrame(
        {
            "id_rna": list(adata_rna.obs_names),
            "rna_idx": np.arange(adata_rna.shape[0]).astype(int),
        }
    )
    df_acc = df_acc.merge(df_RNA, on="id_rna")

    # extract a common set of cells
    sel_idx = np.array(df_acc["rna_idx"]).astype(int)
    adata_rna_v1 = adata_rna[sel_idx]
    sel_idx = np.array(df_acc["acc_idx"]).astype(int)
    adata_acc_v1 = adata_acc_v0[sel_idx]

    # transfer label, since they now share a common index (rna_id)
    adata_acc_v1.obs["state_info"] = adata_rna_v1.obs["state_info"]
    adata_acc_v1.obs["fate_bias"] = adata_rna_v1.obs["fate_bias"]
    adata_acc_v1.obs["fate_map"] = adata_rna_v1.obs["fate_map"]
    adata_acc_v1.uns["fate_map_colors"] = adata_rna_v1.uns["fate_map_colors"]
    adata_acc_v1.uns["state_info_colors"] = adata_rna_v1.uns["fate_map_colors"]
    adata_acc_v1.obsm["X_umap_rna"] = adata_rna_v1.obsm["X_umap"]

    sc.pl.embedding(adata_acc_v1, color="fate_map", basis="X_umap", s=200)
    return adata_acc_v1


def merge_pseudo_bulk(
    data_path,
    groupby_key="celltype",
    selected_fates=None,
    source="met",
    save_data_des=None,
):
    """
    selected_fates is the names of precomputed bulk files that will be aggregated together. If selected_fates are not provided, using groupby_key to automatically
    determine selected_fates

    save_data_des: the file name to save the data. If not provided, it will be the same as groupby_key
    """

    if source == "met":
        source_dir = "met/cpg_level"
    else:
        source_dir = "acc/gpc_level"

    df_sample = load_sample_info(data_path)

    if selected_fates is None:
        selected_fates = df_sample[groupby_key].dropna().unique()
        print(f"Selected cluster names: {selected_fates}")
    df_1 = pd.read_csv(
        f"{data_path}/{source_dir}/pseudobulk/{groupby_key}/{selected_fates[0]}.tsv.gz",
        sep="\t",
        compression="gzip",
    )

    if save_data_des is None:
        save_data_des = groupby_key
    file_out = f"{data_path}/{source_dir}/pseudobulk/{save_data_des}.tsv.gz"
    print(f"The results will be saved at {file_out}")

    for x in selected_fates[1:]:
        print(f"Current sample: {x}")
        df_2 = pd.read_csv(
            f"{data_path}/{source_dir}/pseudobulk/{groupby_key}/{x}.tsv.gz",
            sep="\t",
            compression="gzip",
        )
        df_merge = df_1.merge(df_2, on=["chr", "pos"], how="outer").fillna(0)
        df_merge["met_sites"] = df_merge["met_sites_x"] + df_merge["met_sites_y"]
        df_merge["nonmet_sites"] = (
            df_merge["nonmet_sites_x"] + df_merge["nonmet_sites_y"]
        )

        df_merge["met_sites"] = df_merge["met_sites"].astype(int)
        df_merge["nonmet_sites"] = df_merge["nonmet_sites"].astype(int)

        df_merge["rate"] = (
            100
            * df_merge["met_sites"]
            / (df_merge["met_sites"] + df_merge["nonmet_sites"])
        )

        df_merge = df_merge.dropna()

        df_merge["rate"] = (df_merge["rate"]).astype(int)
        df_1 = df_merge.filter(["chr", "pos", "met_sites", "nonmet_sites", "rate"])

    print("save data")
    df_1.to_csv(
        file_out,
        sep="\t",
        compression="gzip",
        index=0,
    )



import time

import numpy as np
import scipy.sparse as ssp
from matplotlib import pyplot as plt
from tqdm import tqdm

from .settings import *


def compute_similarity_matrix(
    input_X_orig,
    method="correlation",
    cores=64,
):
    """
    Calculate the similarity matrix by explicitly excluding entries with NAN

    When computing similarity between two cells, only use the features detected
    in both cells to calculate the result.

    Parameters
    ---------
        input_X:
            cell (N) by feature (M) matrix, a numpy array
        method:
            Method to compute the similarity. Currently only support "correlation", "correlation_fast".
            correlation_fast is also an exact method.

    Returns
    -------
        corr:
            The similarity matrix

    """
    print(f"Use {method} for similarity")
    start = time.time()
    from scipy.stats import spearmanr

    def spearman_corr(X, Y):
        correlation, p_value = spearmanr(X, Y)
        return correlation

    def pearson_corr(X, Y):
        return np.corrcoef(X, Y)[0, 1]

    def jaccard_similarity(X, Y):
        matches = np.sum(X == Y)
        return matches / len(X)

    def manhattan_similarity(X, Y):
        return 1 - np.sum(np.abs(X - Y)) / len(X)

    ## Most of these methods only support binary data format. Also, these methods are not compatible with our similarity correction method
    method_dict = {
        "correlation": pearson_corr,
        "jaccard": jaccard_similarity,
        "manhattan": manhattan_similarity,
        "spearman": spearman_corr,
    }

    N = input_X_orig.shape[0]
    M = input_X_orig.shape[1]
    shared_site_matrix = np.zeros((N, N))

    if method != "correlation_fast":
        # use a soft-link, because we will not modify input_X in this method
        if method not in method_dict.keys():
            raise ValueError(
                f"method should be among {method_dict.keys()}, but has the value {method}"
            )

        input_X = input_X_orig

        non_nan_matrix = ~np.isnan(input_X)
        # X_similarity = np.zeros((N, N))

        def compute_similarity_core(i):
            X_similarity_vector = np.zeros(N)
            shared_site_matrix_vector = np.zeros(N)
            for j in range(i + 1, N):
                joint_non_nan_index = non_nan_matrix[i] & non_nan_matrix[j]
                X = input_X[i][joint_non_nan_index]
                Y = input_X[j][joint_non_nan_index]
                X_similarity_vector[j] = method_dict[method](X, Y)
                shared_site_matrix_vector[j] = np.sum(joint_non_nan_index)
            return X_similarity_vector, shared_site_matrix_vector

        from joblib import Parallel, delayed

        results = Parallel(n_jobs=cores)(
            delayed(compute_similarity_core)(_) for _ in range(N)
        )
        X_similarity = np.array([x[0] for x in results])
        shared_site_matrix = np.array([x[1] for x in results])

        X_similarity = X_similarity + X_similarity.T
        diag_index = np.diag(np.ones(X_similarity.shape[0])) > 0
        X_similarity[diag_index] = 1
        mask = np.zeros_like(X_similarity, dtype=bool)
        mask[np.triu_indices_from(X_similarity, k=1)] = True

        NaN_num = np.sum(np.isnan(X_similarity))
        if NaN_num > 0:
            print(
                f"Warn: {NaN_num} NaN generated in X_similarity, filled with global mean"
            )
            tru_mean = X_similarity[~np.isnan(X_similarity)].mean()
            X_similarity = np.nan_to_num(X_similarity, nan=tru_mean)

    elif method == "correlation_fast":
        # make a copy so that we can modify this input_X
        if ssp.issparse(input_X_orig):
            input_X = input_X_orig.A.copy()
        else:
            input_X = input_X_orig.copy()

        non_nan_matrix = ~np.isnan(input_X)
        non_nan_count = (~non_nan_matrix).sum()

        if non_nan_count > 0:
            # after zero-centering, correlation becomes cosine similarity.
            # Note that here we use the mean of the entire vector from a cell, rather than using
            # only the shared region with another cell j to compute the mean.
            # This is the only difference with the exact/slow method
            # the assumption is that this global mean and local mean within the shared regions should be rather similar
            print("-------zero centered--------")
            for x in range(N):
                tmp = input_X[x]
                input_X[x] = tmp - np.mean(tmp[~np.isnan(tmp)])

            input_X = np.nan_to_num(
                input_X, nan=0
            )  # by replacing nan with zero, the dot product below only considers values with non-zero entries between two cells.
            cosine_dot_product = input_X.dot(input_X.T)

            # compute the pair-wise normalization factor. Need to consider regions detected in both cells
            X_square = input_X * input_X  # used for normalization
            norm_array = np.ones((N, N))

            for i in tqdm(range(N)):
                for j in range(i + 1, N):
                    # use the boolean vector, and summation. This is computationally cheaper than multiplication
                    joint_non_nan_index = non_nan_matrix[i] & non_nan_matrix[j]
                    norm_array[i, j] = np.sum(X_square[i][joint_non_nan_index])
                    norm_array[j, i] = np.sum(X_square[j][joint_non_nan_index])
                    shared_site_matrix[i, j] = np.sum(joint_non_nan_index)

            X_similarity = cosine_dot_product / np.sqrt(norm_array * norm_array.T)
            diag_index = np.diag(np.ones(X_similarity.shape[0])) > 0
            X_similarity[diag_index] = 1
        else:
            X_similarity = fast_pearson_correlation_between_matrix_without_NaN(
                input_X, input_X
            )

    shared_site_matrix = shared_site_matrix + shared_site_matrix.T
    end = time.time()
    print(f"duration: {end-start}")

    return X_similarity, shared_site_matrix


def correct_similarity_matrix(
    X_similarity_old,
    fig_out=True,
    method="fast",
    epsilon=0.01,
    lower_cutoff=0.4,
    upper_cutoff=2.5,
):
    """
    We assume that X_similarity is a square matrix
    """

    if method == "fast":
        print("Use fast/analytical correction method")
    else:
        print("Use old/slow correction method")

    def cost(similarity_matrix):
        tmp = similarity_matrix[np.triu_indices_from(similarity_matrix, k=1)]
        return np.std(tmp) / np.mean(tmp)

    def weight_rescaling(weight):
        # mean-center before thresholding. Rescaling the weight does the change our cost function
        # but it affects the similarity, and improves the final result
        weight = weight / np.mean(weight)
        return weight

    def compute_increment_for_current_cell(j0):
        weight_vector_tmp = weight_vector.copy()
        weight_vector_tmp[j0] = (
            weight_vector_tmp[j0] + epsilon
        )  # *(np.random.randint(2)[0]-1)
        weight_matrix_tmp = (
            weight_vector_tmp[:, np.newaxis] * weight_vector_tmp[np.newaxis, :]
        )

        weighted_X_similarity_tmp = X_similarity_old * weight_matrix_tmp
        std_1d_tmp = cost(weighted_X_similarity_tmp)
        return -(std_1d_tmp - std_1d_new)

    ## initialize the weight vector
    max_values = []
    for i in range(X_similarity_old.shape[0]):
        vector_drop_diag = np.hstack(
            [X_similarity_old[i, :i], X_similarity_old[i, (i + 1) :]]
        )
        max_values.append(abs(vector_drop_diag.max()))

    weight_vector = weight_rescaling(1 / np.sqrt(np.array(max_values)))

    std_history = [1000]
    weight_vector_history = [weight_vector]

    weight_matrix_initial = weight_vector[:, np.newaxis] * weight_vector[np.newaxis, :]
    weighted_X_similarity = X_similarity_old * weight_matrix_initial
    std_1d_new = cost(weighted_X_similarity)
    std_history.append(std_1d_new)
    weight_vector_history.append(weight_vector)

    diff = np.zeros(len(weight_vector))
    for _ in tqdm(range(1000)):
        ## compute the optimal direction of improvement. This mitigates the risk of overfitting
        if method == "fast":
            S_old = X_similarity_old.copy()
            S_new = weighted_X_similarity.copy()
            N0 = S_old.shape[0]
            norm_factor = N0 * (N0 - 1) / 2
            np.fill_diagonal(S_old, np.nan)
            np.fill_diagonal(S_new, np.nan)

            S_new_std = np.nanstd(S_new.flatten())
            S_new_mean = np.nanmean(S_new.flatten())
            diff_X = S_old * (weight_vector[:, np.newaxis])

            dY_dz = np.nansum((S_new - S_new_mean) * (diff_X), axis=0) * 2 / norm_factor
            dX_dz = np.nansum(diff_X, axis=0) / (norm_factor)
            diff_analytic = (
                1 / (2 * S_new_std * S_new_mean) * dY_dz
                - S_new_std / (S_new_mean * S_new_mean) * dX_dz
            )
            diff = -diff_analytic
        else:
            for j in np.arange(len(weight_vector)):
                diff[j] = compute_increment_for_current_cell(j)

        ## update the weight vector, and compute the new similarity
        weight_vector = weight_rescaling(
            weight_vector + epsilon * diff / np.max(abs(diff))
        )
        weight_vector_history.append(weight_vector)

        weight_matrix = weight_vector[:, np.newaxis] * weight_vector[np.newaxis, :]
        weighted_X_similarity = X_similarity_old * weight_matrix
        std_1d_new = cost(weighted_X_similarity)
        std_history.append(std_1d_new)

        ## check convergence and then exit
        excess_weight = (np.sum(weight_vector >= upper_cutoff) > 0) or (
            np.sum(weight_vector <= lower_cutoff) > 0
        )
        if (std_1d_new >= std_history[-2]) or excess_weight:
            break

    weight_vector = weight_vector_history[np.argmin(std_history)]
    weight_matrix = weight_vector[:, np.newaxis] * weight_vector[np.newaxis, :]
    weight_X_similarity = X_similarity_old * weight_matrix
    X_similarity = weight_X_similarity

    if fig_out:
        plt.subplots(figsize=(4, 3.5))
        plt.plot(
            np.arange(len(weight_vector)),
            weight_vector_history[0],
            ".",
            label="initial",
        )
        plt.plot(np.arange(len(weight_vector)), weight_vector, ".", label="final")
        plt.ylim([0, 2])
        plt.xlabel("Cell order")
        plt.ylabel("Weight")
        plt.legend()
        plt.show()

        # std_history = np.array(std_history)
        # plt.subplots(figsize=(4, 3.5))
        # plt.plot(np.arange(len(std_history) - 1), std_history[1:], ".")
        # plt.xlabel("Iteration")
        # plt.ylabel("Std/mean")

        off_diag_idx = ~(np.diag(np.ones(X_similarity_old.shape[0])) > 0)
        temp_X_orig = weight_matrix_initial[off_diag_idx].flatten()
        if len(temp_X_orig) > 5000:
            sel_idx = np.random.choice(np.arange(len(temp_X_orig)), 5000)
        else:
            sel_idx = np.arange(len(temp_X_orig))
        plt.subplots(figsize=(4, 3.5))
        plt.plot(
            temp_X_orig[sel_idx],
            X_similarity_old[off_diag_idx].flatten()[sel_idx],
            ".b",
        )
        plt.xlabel("1/sqrt(max_i*max_j)")
        plt.ylabel("Similarity")
        plt.title("Original")
        plt.show()

        # plt.subplots(figsize=(4, 3.5))
        # plt.plot(
        #     temp_X_orig[sel_idx],
        #     X_similarity[off_diag_idx].flatten()[sel_idx],
        #     ".r",
        # )
        # plt.xlabel("1/sqrt(max_i*max_j)")
        # plt.ylabel("Similarity")
        # plt.title("New")
        # plt.show()

    return X_similarity


def remove_celltype_signal_discrete_state(
    X_similarity,
    cell_type_array,
    outlier_threshold=1,
):
    """
    Cell-specific zscore.

    S=T+L
    Here, we calcualte the cell-specific T_ij.  Assume that i belongs to type p, and j to q.
    mean_T_i: mean similarity between i and all cells with type q (from j)
    mean_T_j: mean similarity between j and all cells with type p (from i)
    Then, compute the mean: T_ij=(mean_T_i+mean_T_j)/2

    This method proves to be more accurate than just using the mean similarity
    between all cells from type p and type q.

    outlier_threshold=1
    A smaller outlier_threshold excludes more cells.
    A very large outlier_threshold means that nothing is excluded.

    Remove the diagonal effect, and the outliers, proves to be useful.
    """

    unique_cell_types = list(set(cell_type_array))

    mean_matrix = np.zeros(X_similarity.shape)
    all_values = X_similarity[np.triu_indices(X_similarity.shape[0], k=1)]
    global_mean = np.mean(all_values)

    for i in range(len(unique_cell_types)):
        for j in range(i, len(unique_cell_types)):
            x = unique_cell_types[i]
            y = unique_cell_types[j]

            sel_index_1 = np.nonzero(cell_type_array == x)[0]
            sel_index_2 = np.nonzero(cell_type_array == y)[0]

            X_sub = X_similarity[sel_index_1][:, sel_index_2].copy()
            values = X_sub.flatten()

            # remove cell-pairs with abnormally high similarity (due to lineage similarity)
            # This improves the result.
            cutoff_tmp_high = np.mean(values) + outlier_threshold * np.std(values)
            X_sub[(X_sub > cutoff_tmp_high)] = np.nan

            if i == j:
                np.fill_diagonal(X_sub, np.nan)
                X0 = np.nanmean(X_sub, axis=0, keepdims=True)
                X_mean = (X0 + X0.T) / 2
            else:
                X0 = np.nanmean(X_sub, axis=0, keepdims=True)
                X1 = np.nanmean(X_sub, axis=1, keepdims=True)
                X_mean = (X0 + X1) / 2

            mean_matrix[np.ix_(sel_index_1, sel_index_2)] = X_mean
            mean_matrix[np.ix_(sel_index_2, sel_index_1)] = X_mean.T

    mean_matrix[np.isnan(mean_matrix)] = global_mean
    new_S_matrix = X_similarity - mean_matrix
    np.fill_diagonal(new_S_matrix, 1)

    return new_S_matrix


def remove_celltype_signal_continuous_state(
    X_similarity, neighbor_info, outlier_threshold=1
):
    all_values = X_similarity[np.triu_indices(X_similarity.shape[0], k=1)]
    global_mean_x = np.mean(all_values)
    mean_matrix = np.zeros(X_similarity.shape)

    for i in range(X_similarity.shape[0]):
        for j in range(i, X_similarity.shape[0]):
            index_list = [i, j]
            mean_value_list = []
            for _ in index_list:
                sel_index_1 = np.array(neighbor_info[_])
                if i == j:
                    sel_index_2 = sel_index_1
                    # This part is different from the fast method. But it only affects the diagonal
                    # And the diagonal is eventually set to be 1
                else:
                    tmp = index_list.copy()
                    tmp.remove(_)
                    sel_index_2 = [tmp[0]]

                if _ == i:
                    X_sub = X_similarity[sel_index_1][:, sel_index_2]
                else:
                    X_sub = X_similarity[sel_index_2][:, sel_index_1]

                if i == j:
                    values = X_sub[np.triu_indices(X_sub.shape[0], k=1)]
                else:
                    values = X_sub.flatten()

                # remove cell-pairs with abnormally high similarity (due to lineage similarity)
                cutoff_tmp_high = np.mean(values) + outlier_threshold * np.std(values)
                mean_value = np.mean(values[(values < cutoff_tmp_high)])
                mean_value_list.append(mean_value)

            mean_matrix[i, j] = np.mean(mean_value_list)
            mean_matrix[j, i] = mean_matrix[i, j]

    mean_matrix[np.isnan(mean_matrix)] = global_mean_x
    new_S_matrix = X_similarity - mean_matrix
    np.fill_diagonal(new_S_matrix, 1)
    return new_S_matrix


def rescale_similarity_matrix(X_similarity):
    temp_array = X_similarity[np.triu_indices_from(X_similarity, k=1)]
    min_value = temp_array.min()
    max_value = temp_array.max()
    # add a max weight 0.999, so that only the diagonal will be 1
    X_similarity_new = 0.999 * (X_similarity - min_value) / (max_value - min_value)
    np.fill_diagonal(X_similarity_new, 1)
    return X_similarity_new


def fast_pearson_correlation_between_matrix_without_NaN(x, y):
    """
    compute the correlation along the axis=1.
    So, if x has the shape: n1*M
           y has the shape: n2:M
        Then, the result has the shape: n1*n2
    """
    x1 = x.transpose()
    y1 = y.transpose()
    xv = x1 - x1.mean(axis=0)
    yv = y1 - y1.mean(axis=0)
    xvss = (xv * xv).sum(axis=0)
    yvss = (yv * yv).sum(axis=0)
    result = np.matmul(xv.transpose(), yv) / np.sqrt(np.outer(xvss, yvss))
    # bound the values to -1 to 1 in the event of precision issues
    return (result + result.T) / 2



from . import QC, adata, analysis, batch, clone
from . import help_functions as hf
from . import (
    lineage,
    metadata,
    metric,
    plotting,
    region,
    scripts,
    settings,
    similarity,
    simulation,
)



import os
import random
import warnings

import cospar as cs
import numpy as np
import pandas as pd
import scanpy as sc
import seaborn as sns
from ete3 import Tree
from matplotlib import pyplot as plt
from skbio import DistanceMatrix

from . import clone, lineage, metadata, metric, plotting, similarity
from .settings import *

warnings.filterwarnings("ignore")
pd.options.mode.chained_assignment = None  # disable the SettingWithCopyWarning


def methylserver_call(
    raw_data_path,
    metadata_path,
    clone_key,
    out_dir,
    data_des="",
    save_data_des="",
    cell_type_key=None,
    heatmap_additional_key_list=None,
    **kwargs,
):
    """
    Used by the methyltree web portal
    """

    df_new2 = pd.read_csv(raw_data_path, compression="gzip", sep="\t").rename(
        columns={"cell_id": "index"}
    )
    df_out = df_new2.pivot(index="index", columns="genomic_region_id", values="value")
    adata = sc.AnnData(df_out)

    df_sample = pd.read_csv(metadata_path, compression="gzip", sep="\t")
    metadata.update_sample_info_on_adata(adata, df_sample)

    if clone_key not in adata.obs.keys():
        raise ValueError(
            f"clone_key={clone_key}, which is not valid. Must be among {adata.obs.keys()}"
        )

    if cell_type_key == "":
        cell_type_key = None
    if cell_type_key is not None:
        if cell_type_key not in adata.obs.keys():
            raise ValueError(
                f"cell_type_key={clone_key}, which is not valid. Must be among {adata.obs.keys()}"
            )
    if heatmap_additional_key_list is not None:
        for _ in heatmap_additional_key_list:
            if _ not in adata.obs.keys():
                raise ValueError(
                    f"heatmap_additional_key_list={heatmap_additional_key_list}, which is not valid. Must be among {adata.obs.keys()}"
                )

    os.makedirs(out_dir, exist_ok=True)

    adata, stat_out = comprehensive_lineage_analysis(
        out_dir,
        data_path=None,
        save_data_des=save_data_des,
        clone_key=clone_key,
        adata_orig=adata,
        perform_accuracy_estimation=True,
        perform_depth_analysis=True,
        perform_memory_analysis=True,
        cell_type_key=cell_type_key,
        heatmap_additional_key_list=heatmap_additional_key_list,
        update_sample_info=False,
        data_des=data_des,
        **kwargs,
    )
    return adata, stat_out


def comprehensive_lineage_analysis(
    out_dir,
    data_path,
    save_data_des,
    clone_key,
    adata_orig=None,
    selected_idx=None,
    update_sample_info=False,
    compute_similarity=False,
    similarity_method="correlation",
    neighbor_info=None,
    cell_type_key=None,
    similarity_correction=True,
    similarity_correction_method="fast",
    similarity_correction_step_size=0.05,
    similarity_normalize=True,
    remove_celltype_signal=False,
    remove_celltype_signal_method="fast",
    remove_celltype_signal_outlier_thresh=1,
    fig_dir=None,
    data_des="",
    fig_out=True,
    heatmap_ordering_method="UPGMA",
    optimize_tree=False,
    optimize_background_cutoff=0.8,
    optimize_show_legend=True,
    clone_color_dict=None,
    heatmap_title=None,
    heatmap_flip_order=False,
    heatmap_show_label=False,
    permute_cell_order=False,
    heatmap_rotate_cells=False,
    heatmap_vmin_percentile=50,
    heatmap_vmax_percentile=99,
    heatmap_maximum_labels_per_column=40,
    heatmap_fontsize=4,
    heatmap_figsize=(10, 9.5),
    heatmap_show_legend=True,
    heatmap_additional_key_list=None,
    heatmap_max_clone_N=22,
    heatmap_min_clone_size=2,
    perform_clone_inference=False,
    clone_inference_threshold=0.5,
    clone_inference_signal_noise_plot=True,
    clone_inference_print=False,
    perform_coarse_graining=False,
    coarse_grain_figsize=(4, 3.5),
    coarse_grain_vmin=None,
    coarse_grain_normalization=True,
    perform_accuracy_estimation=True,
    perform_memory_analysis=False,
    perform_depth_analysis=False,
    accuracy_key="accuracy",
    save_adata=False,
):
    """
    We use this function to perform systematic lineage analysis on DNA methylation data.

    Module:
    1, Data loading
    2, X_similarity regression to remove noise
        * similarity_correction_std=False,
        * similarity_correction_std_per_celltype=False,
        * similarity_correction_threshold=None,
    3, X_similarity z-score to enhance lineage signal
        * remove_celltype_signal=False,
    4, Lineage tree inference or cell state ordering
        * order_method=['NJ','UPGMA','HierarchicalClustering']
    5, Improve lineage tree with stochastic optimization
        * optimize_tree=False,
        * optimize_background_cutoff=0.8,
    6, Identify putative clones with thresholding
        * identify_clones=False,
        * InfClone_threshold=0.5,
        * InfClone_signal_noise=True,
    7, Clean up ordered_clone_array and compute lineage accuracy
    8, Visualize the X_similarity
        * replot_similarity_matrix=False
    9, Memory analysis
    10, Coarse-grain analysis
    11, Accuracy by depth

    ----------------
    If you want to specify colors for each clone, you can provide clone_color_dict
    """

    #################################
    ##        Data loading
    #################################

    if fig_dir is None:
        fig_dir = f"{out_dir}/figure"
        os.makedirs(fig_dir, exist_ok=True)
    sc.set_figure_params()
    sc.settings.figdir = fig_dir

    if adata_orig is None:
        adata = sc.read(f"{out_dir}/{save_data_des}.h5ad")
    else:
        adata = adata_orig.copy()
        print("use provided adata")
    print("adata shape: ", adata.shape)

    ##########################################
    ## check or update metadata
    ##########################################

    # check if df_sample needs to be updated
    used_key_list = [clone_key, cell_type_key]
    if heatmap_additional_key_list is not None:
        used_key_list = used_key_list + heatmap_additional_key_list
    for _ in used_key_list:
        if _ is not None:
            if _ not in adata.obs.keys():
                df_sample = metadata.load_sample_info(data_path)
                if _ not in df_sample.columns:
                    available_keys = set(adata.obs.keys()).union(set(df_sample.columns))
                    raise ValueError(
                        f"The selected key '{_}' it not available. \nPlease select from {available_keys}"
                    )
                else:
                    update_sample_info = True
                    print(
                        f"Warning: '{_}' is not a valid key. Force to update df_sample"
                    )

    if update_sample_info:
        print("update sample")
        df_sample = metadata.load_sample_info(data_path)
        metadata.update_sample_info_on_adata(adata, df_sample)

    if f"X_similarity_{similarity_method}_raw" not in adata.obsm.keys():
        compute_similarity = True
        if adata_orig is None:
            save_adata = True

    adata, order_x = lineage.methyltree_core(
        adata,
        out_dir,
        save_data_des,
        clone_key,
        compute_similarity=compute_similarity,
        similarity_method=similarity_method,
        similarity_correction=similarity_correction,
        similarity_correction_method=similarity_correction_method,
        similarity_correction_step_size=similarity_correction_step_size,
        similarity_normalize=similarity_normalize,
        remove_celltype_signal=remove_celltype_signal,
        remove_celltype_signal_method=remove_celltype_signal_method,
        remove_celltype_signal_outlier_thresh=remove_celltype_signal_outlier_thresh,
        cell_type_key=cell_type_key,
        neighbor_info=neighbor_info,
        heatmap_ordering_method=heatmap_ordering_method,
        optimize_tree=optimize_tree,
        optimize_background_cutoff=optimize_background_cutoff,
        optimize_show_legend=optimize_show_legend,
        permute_cell_order=permute_cell_order,
        selected_idx=selected_idx,
        fig_out=fig_out,
    )

    X_similarity = adata.obsm["X_similarity"]
    with open(f"{out_dir}/lineage_tree_{clone_key}_{save_data_des}.txt", "r") as f:
        my_tree = Tree(f.read())

    ###############################################
    ## Identify putative clones with thresholding
    ###############################################
    adata.obs["state_info_idx"] = np.arange(adata.shape[0]).astype(str)
    df_state_clone = pd.DataFrame(
        {
            "state_info": adata.obs["state_info_idx"],
            "clone_id": adata.obs[clone_key].to_list(),
        }
    ).sort_values("clone_id")
    df_state_clone.to_csv(f"{out_dir}/state_clone_info_{clone_key}_{save_data_des}.csv")

    if perform_clone_inference:
        rename_leaf_map = dict(
            zip(df_state_clone["state_info"].astype(str), df_state_clone["clone_id"])
        )
        df_tmp = clone.identify_putative_clones(
            X_similarity,
            my_tree,
            weight_threshold=clone_inference_threshold,
            plot_signal_noise=clone_inference_signal_noise_plot,
            print_clone=clone_inference_print,
            rename_leaf_map=rename_leaf_map,
        )

        inferred_clone_N = len(df_tmp["predicted_clone"].unique())
        print(f"Inferred clone number: {inferred_clone_N}")

        adata.obs["inferred_clones"] = df_tmp.sort_values("cell_id")[
            "predicted_clone"
        ].to_numpy()

        sel_idx = pd.isna(adata.obs["inferred_clones"])
        old_clone_array = adata.obs[clone_key].copy().to_numpy()
        old_clone_array[sel_idx] = np.nan
        # switch to new clone_key
        clone_key = "filtered_clone"
        adata.obs[clone_key] = old_clone_array

    ############################################################
    ## Clean up ordered_clone_array and compute lineage accuracy
    ############################################################

    if heatmap_flip_order:
        order_x = np.array(order_x)[::-1]

    ordered_clone_array = np.array(adata.obs[clone_key])[order_x].astype(str)
    unique_clone_set = list(set(ordered_clone_array))

    ## rotate the vector so that the first clone will be put at the end
    ## this is because our algorithm often separates cells from the same clone at the beginning and end

    if heatmap_rotate_cells:
        x0 = ordered_clone_array[0]
        count_0 = 0
        unique_clone_set = list(set(ordered_clone_array))
        if "nan" in unique_clone_set:
            unique_clone_set.remove("nan")

        if len(unique_clone_set) > 1:
            while (ordered_clone_array[count_0] == x0) or (
                ordered_clone_array[count_0] == "nan"
            ):
                count_0 += 1
            ordered_clone_array = np.hstack(
                [
                    ordered_clone_array[count_0:],
                    np.array(ordered_clone_array[:count_0])[::-1],
                ]
            )
            order_x = np.hstack([order_x[count_0:], np.array(order_x[:count_0])[::-1]])

    adata.uns["order_x"] = order_x
    adata.uns["ordered_clone_array"] = ordered_clone_array

    accuracy_stat = {}
    if perform_accuracy_estimation:
        if heatmap_min_clone_size > 1:
            ## exluding these signleton clones will artificially inflate the accuracy. [clone_1,clone_0,clone_1] becomes [clone_1,NAN,clone_1]
            for x0 in unique_clone_set:
                if np.sum(ordered_clone_array == x0) == 1:
                    ordered_clone_array[ordered_clone_array == x0] = "nan"

        np.save(
            f"{out_dir}/lineage_order_{clone_key}_{save_data_des}.npy",
            ordered_clone_array,
        )

        # only consider clones with at least two cells before calculating accuracy
        df_accuracy = metric.lineage_accuracy_from_leaf_order(ordered_clone_array)

        # random accuracy
        num_iterations = 10
        random_accuracies = []
        for _ in range(num_iterations):
            random_clone_array = ordered_clone_array.copy()
            random.shuffle(random_clone_array)
            df_random_accuracy = metric.lineage_accuracy_from_leaf_order(
                random_clone_array
            )
            random_accuracies.append(df_random_accuracy)
        df_random_accuracies_concat = pd.concat(random_accuracies)
        df_random_accuracy = (
            df_random_accuracies_concat.groupby(["clone", "clone_size"])
            .mean()
            .reset_index()
        )
        columns_to_prefix = ["continuity", "accuracy", "entropy", "wassertein"]
        df_random_accuracy.rename(
            columns={col: f"random_{col}" for col in columns_to_prefix}, inplace=True
        )

        df_accuracy = df_accuracy.merge(
            df_random_accuracy, on=["clone", "clone_size"]
        ).sort_values(by="clone")
        print(df_accuracy.head(5))

        for sel_key in [
            "continuity",
            "accuracy",
            "entropy",
            "wassertein",
            "random_continuity",
            "random_accuracy",
            "random_entropy",
            "random_wassertein",
        ]:
            weighted_accuracy = np.sum(
                df_accuracy[sel_key]
                * df_accuracy["clone_size"]
                / df_accuracy["clone_size"].sum()
            )
            mean_accuracy = df_accuracy[sel_key].mean()
            accuracy_stat[f"weighted_{sel_key}"] = weighted_accuracy
            accuracy_stat[f"mean_{sel_key}"] = mean_accuracy

            # print(f"weighted {sel_key}:", weighted_accuracy)
            print(f"mean {sel_key}:", mean_accuracy)

        df_accuracy.to_csv(
            f"{out_dir}/accuracy_{clone_key}_{save_data_des}.csv", index=0
        )
        adata.uns["accuracy"] = accuracy_stat

    #################################################################
    ##  visualize the X_similarity
    #################################################################

    plotting.plot_similarity_heatmap_with_multiple_colorbars(
        adata,
        additional_key_list=heatmap_additional_key_list,
        heatmap_vmin_percentile=heatmap_vmin_percentile,
        heatmap_vmax_percentile=heatmap_vmax_percentile,
        figsize=heatmap_figsize,
        fig_dir=fig_dir,
        data_des=data_des,
        show_label=heatmap_show_label,
        title=heatmap_title,
        rotate_cells=heatmap_rotate_cells,
        clone_color_dict_orig=clone_color_dict,
        fontsize=heatmap_fontsize,
        show_legend=heatmap_show_legend,
        maximum_labels_per_column=heatmap_maximum_labels_per_column,
        max_clone_N=heatmap_max_clone_N,
        min_clone_size=heatmap_min_clone_size,
    )

    ###################
    ## Memory analysis
    ###################

    if perform_memory_analysis:
        background, mean_similarity_list, pvalue_all = memory_analysis(
            adata.obsm["X_similarity"][order_x][:, order_x],
            ordered_clone_array,
            fig_dir=fig_dir,
            save_data_des=data_des,
        )

        obs_mean = np.mean(mean_similarity_list)
        rand_mean = np.mean(background)
        rand_std = np.std(background)
        memory_strength = (obs_mean - rand_mean) / rand_std
        clone_N = len(set([x for x in ordered_clone_array if x != "nan"]))
        accuracy_stat["memory_strength"] = memory_strength
        accuracy_stat["pvalue"] = pvalue_all
        accuracy_stat["clone_N"] = clone_N

    ##############################
    ## Coarse-grain analysis
    ##############################

    if perform_coarse_graining:
        adata.uns["data_des"] = [data_des]
        coarse_grain_analysis(
            adata,
            clone_key,
            vmin=coarse_grain_vmin,
            figsize=coarse_grain_figsize,
            method=heatmap_ordering_method,
            title=heatmap_title,
            normalized_X_similarity=coarse_grain_normalization,
        )

    #########################
    # Accuracy by tree depth
    #########################
    if perform_depth_analysis:
        if heatmap_ordering_method in ["NJ", "UPGMA"]:
            ###############################################
            ## Accuracy at different clone threshold
            ###############################################
            cs.settings.set_figure_params()
            df_stats = metric.accuracy_at_different_similarity_threshold(
                adata.obsm["X_similarity"],
                my_tree,
                np.array(adata.obs[clone_key]),
                accuracy_key=accuracy_key,
            )
            df_stats.to_csv(
                f"{out_dir}/accuracy_at_varying_threshold_{clone_key}_{save_data_des}.csv",
                index=0,
            )

            ### Accuracy by tree depth
            df_accuracy_by_depth = metric.compute_accuracy_by_depth(
                my_tree, (adata.obs[clone_key]), accuracy_key=accuracy_key
            )

            df_plot = (
                df_accuracy_by_depth.groupby("Depth")
                .agg(
                    accuracy=(accuracy_key, "mean"),
                    clone=("clone", lambda x: "@".join(x)),
                )
                .reset_index()
            )
            df_plot["clone_N"] = df_plot["clone"].apply(
                lambda x: len(set(x.split("@")))
            )
            # display(df_plot)
            from matplotlib import rcParams

            color_1 = "#d7301f"
            color_2 = "#225ea8"
            cs.settings.set_figure_params()
            rcParams["axes.spines.right"] = True
            fig, ax = plt.subplots(figsize=(5, 4))
            ax.plot(
                df_plot["Depth"],
                df_plot["accuracy"],
                "o-",
                color=color_1,
            )
            ax2 = plt.twinx()
            ax2.plot(df_plot["Depth"], df_plot["clone_N"], "--", color=color_2)
            ax.set_xlabel("Lineage tree depth")
            ax.set_ylim([0, 1.03])
            ax.set_ylabel(f"Reconstruction {accuracy_key}", color=color_1)
            ax2.set_ylabel("Clone number (size>1)", color=color_2)
            plt.tight_layout()
            rcParams["axes.spines.right"] = False
            plt.savefig(f"{fig_dir}/{data_des}_accuracy_by_depth.pdf")

    adata_path = f"{out_dir}/{save_data_des}.h5ad"
    if save_adata:
        print(f"save data at ")
        print(adata_path)
        selected_fatures = ["sample"] + used_key_list
        if update_sample_info:
            df_sample = metadata.load_sample_info(data_path)
            for x in list(set(df_sample.columns).difference(selected_fatures)):
                del adata.obs[x]
        adata.write(adata_path)

    return adata, accuracy_stat


def estimate_transition_rate(seq_1, seq_2):
    """
    transition_rate=[ [0->0, 0->0.5, 0->1],
                      [0.5->0,0.5->0.5,0.5->1],
                      [1->0,1->0.5,1->1] ]
    """

    transition_rate = np.zeros((3, 3))

    joint_non_nan_index = (~np.isnan(seq_1)) & (~np.isnan(seq_2))
    X = seq_1[joint_non_nan_index]
    Y = seq_2[joint_non_nan_index]
    denominator = []

    for i, x in enumerate([0, 0.5, 1]):
        denominator.append(np.sum(X == x))
        for j, y in enumerate([0, 0.5, 1]):
            transition_rate[i, j] = np.sum((X == x) & (Y == y)) / np.sum((X == x))
    print(f"x=0:{denominator[0]};  x=0.5:{denominator[1]}; x=1:{denominator[2]}")
    print(f"mean_rate_x={np.mean(X)};  mean_rate_y={np.mean(Y)}")
    return transition_rate


def coarse_grain_analysis(
    adata,
    clone_key,
    excluded_states=None,
    vmin=None,
    figsize=(4, 3.5),
    method="UPGMA",
    title="",
    semi_triangular=False,
    min_cell_N_per_group=3,
    normalized_X_similarity=True,
):
    cs.settings.set_figure_params(figsize=figsize, format="pdf")
    adata.obs["state_info"] = adata.obs[clone_key]
    df_count = (
        pd.DataFrame({clone_key: adata.obs[clone_key].to_list()})
        .reset_index()
        .groupby(clone_key)
        .agg({"index": "count"})
        .reset_index()
    )
    excluded_states_0 = list(
        df_count[df_count["index"] < min_cell_N_per_group][clone_key]
    )
    if len(excluded_states_0) > 0:
        print(f"Exclude states with <{min_cell_N_per_group} cells: ", excluded_states_0)

    if excluded_states is None:
        excluded_states = excluded_states_0
    else:
        excluded_states = list(set(excluded_states).union(excluded_states_0))
    cs.pp.initialize_adata_object(adata)
    exclude_idx = pd.isna(adata.obs["state_info"]) | np.array(
        adata.obs["state_info"].isin(excluded_states)
    )
    adata_sub = adata[~exclude_idx]
    adata_sub.obsm["X_similarity"] = adata_sub.obsm["X_similarity"][:, ~exclude_idx]

    data_des_0 = adata_sub.uns["data_des"][0]
    adata_sub.uns["data_des"] = [f"{data_des_0}_{clone_key}"]
    if len(set(adata_sub.obs["state_info"])) > 2:
        cs.tl.fate_coupling(
            adata_sub,
            source="X_similarity",
            normalized_X_similarity=normalized_X_similarity,
        )
        cs.pl.fate_coupling(
            adata_sub,
            source="X_similarity",
            vmin=vmin,
            title=title,
            color_bar_label="Methylation similarity score",
            semi_triangular=semi_triangular,
        )

        result_dict = adata_sub.uns[f"fate_coupling_X_similarity"]
        X_coupling = result_dict["X_coupling"]
        fate_names = result_dict["fate_names"]
        dissimilarity = np.max(X_coupling) - X_coupling
        dissimilarity = dissimilarity - np.min(dissimilarity)
        for j in range(X_coupling.shape[0]):
            dissimilarity[j, j] = 0
        dissimilarity = dissimilarity / np.max(dissimilarity)

        dm = DistanceMatrix(
            dissimilarity,
            [f"{x}" for x in list(np.arange(dissimilarity.shape[0]).astype(str))],
        )

        if method == "UPGMA":
            from biotite.sequence.phylo import upgma

            my_tree = Tree(str(upgma(dissimilarity).to_newick()))
        else:
            from skbio.tree import nj

            my_tree = Tree(str(nj(dm)))
            # from . import lineage_tree
            # my_tree = Tree(str(lineage_tree.nj(dm)))

        name_dict = dict(
            zip(np.arange(len(fate_names)).astype(int).astype(str), fate_names)
        )
        for j, leaf in enumerate(my_tree.iter_leaves()):
            leaf.name = name_dict[leaf.name]
        print(my_tree)
        return my_tree

    else:
        raise ValueError("Must have >=3 cluster to generate trees")


def coarse_grained_coupling(
    adata,
    key_1,
    selected_fates_1=None,
    key_2=None,
    selected_fates_2=None,
    selected_times=None,
):
    """
    Generate coarse-grained coupling from two selected keys
    """

    if key_2 == key_1:
        key_2 = key_1

    adata.obs["state_info"] = adata.obs[key_1]
    cs.pp.initialize_adata_object(adata)
    sel_idx = ~(
        pd.isna(adata.obs["state_info"])
        | pd.isna(adata.obs[key_2])
        | (adata.obs["state_info"] == "nan")
        | (adata.obs[key_2] == "nan")
    )
    adata_sub = adata[sel_idx]
    adata_sub.obsm["X_similarity"] = adata_sub.obsm["X_similarity"][:, sel_idx]

    time_info = np.array(adata_sub.obs["time_info"])
    if selected_times is not None:
        if type(selected_times) is not list:
            selected_times = [selected_times]
    sp_idx_0 = cs.hf.selecting_cells_by_time_points(time_info, selected_times)
    state_annote = adata_sub[sp_idx_0].obs["state_info"]
    (
        mega_cluster_list,
        valid_fate_list,
        fate_array_flat,
        sel_index_list,
    ) = cs.hf.analyze_selected_fates(state_annote, selected_fates_1)

    state_annote_2 = adata_sub[sp_idx_0].obs[key_2]
    (
        mega_cluster_list_2,
        valid_fate_list_2,
        fate_array_flat_2,
        sel_index_list_2,
    ) = cs.hf.analyze_selected_fates(state_annote_2, selected_fates_2)

    Smatrix_0 = adata_sub.obsm["X_similarity"]
    Smatrix = Smatrix_0[sp_idx_0][:, sp_idx_0]
    X_coupling = np.zeros((len(mega_cluster_list), len(mega_cluster_list_2)))
    for j, idx_0 in enumerate(sel_index_list):
        for k, idx_1 in enumerate(sel_index_list_2):
            tmp = Smatrix[idx_0][:, idx_1]
            X_coupling[j, k] = np.median(tmp.flatten())

    return X_coupling, mega_cluster_list, mega_cluster_list_2


def memory_analysis(X_similarity_orig, clone_array, fig_dir=".", save_data_des=""):
    """
    clone_array: clone annotation for each of the element in X_similarity_orig matrix
    """

    clone_array = np.array(clone_array).astype(str)

    X_similarity = X_similarity_orig.copy().astype(float)
    temp_array = X_similarity[np.triu_indices_from(X_similarity, k=1)]
    min_value = temp_array.min()
    max_value = temp_array.max()
    X_similarity = (X_similarity - min_value) / (max_value - min_value)

    for i in range(X_similarity.shape[0]):
        X_similarity[i, i] = np.nan

    # scaled_similarity = X_similarity_orig.copy()
    # scaled_similarity[scaled_similarity > 0.99] = np.mean(scaled_similarity)
    # s_max = np.max(scaled_similarity)
    # s_min = np.min(scaled_similarity)
    # scaled_similarity = (scaled_similarity - s_min) / (s_max - s_min)

    clone_key = "clone"
    df_clone_1 = (
        pd.DataFrame({clone_key: clone_array})
        .reset_index()
        .groupby(clone_key)
        .agg({"index": "count"})
    )
    df_clone_1 = (
        df_clone_1[df_clone_1["index"] > 1]
        .sort_values("index", ascending=True)
        .reset_index()
    )

    all_clone_IDs = df_clone_1[clone_key].astype(str).values
    valid_clone_IDs = [x for x in all_clone_IDs if not x.startswith("nan")]

    # to obtain the background, only use the clones that we know, and set the block to be nan (the remaining will be background)
    clone_similarity_score = {}
    for x in valid_clone_IDs:
        sp_idx = np.nonzero(clone_array == x)[0]
        tmp = X_similarity[sp_idx][:, sp_idx].flatten()
        clone_similarity_score[x] = tmp[~np.isnan(tmp)]
        for i in sp_idx:
            for j in sp_idx:
                X_similarity[i, j] = np.nan

    tmp1 = X_similarity.flatten()
    background = tmp1[~np.isnan(tmp1)]

    import scipy.stats as stats

    # here, we only compute the pvalue for valid clones, exluding nan
    pvalue_list = []
    mean_similarity_list = []
    for x in valid_clone_IDs:
        mean_similarity_list.append(np.mean(clone_similarity_score[x]))
        stat_score, pvalue = stats.ranksums(
            clone_similarity_score[x], background, alternative="greater"
        )
        pvalue_list.append(pvalue)

    import statsmodels.sandbox.stats.multicomp

    # P_value = pvalue_list
    # correction assumes that some of these Pvalues are due to randomness. This may not be true.
    P_value = statsmodels.sandbox.stats.multicomp.multipletests(
        list(pvalue_list), alpha=0.05, method="fdr_bh"
    )[1]

    cs.settings.set_figure_params(pointsize=10, fontsize=14)

    plt.subplots(1, 1, figsize=(4, 8))
    x_tick_index = np.linspace(1, len(pvalue_list), len(pvalue_list))
    plt.barh(x_tick_index[::-1], -np.log10(P_value), color="grey")
    plt.yticks(x_tick_index[::-1], np.array(valid_clone_IDs), rotation=0)
    # Set text labels and properties.
    plt.plot(
        np.zeros(len(x_tick_index)) - np.log10(0.05),
        x_tick_index,
        "--",
        color="#a50f15",
    )
    plt.xlabel("-log10 (Pvalue)")
    plt.tight_layout()
    plt.savefig(f"{fig_dir}/{save_data_des}_pvalue_for_individual_clone.pdf")

    ########## all together
    stat_score, pvalue_all = stats.ranksums(
        mean_similarity_list, background, alternative="greater"
    )

    plt.subplots(1, 1, figsize=(5.5, 3.5))
    color_random = "#a6bddb"
    color_data = "#fdbb84"

    all_data = list(mean_similarity_list) + list(background)
    bins = np.linspace(np.min(all_data), np.max(all_data), 50)
    ax = sns.histplot(
        data=background,
        label="Inter-clone",
        bins=bins,
        stat="probability",
        color=color_random,
    )
    ax = sns.histplot(
        data=mean_similarity_list,
        label="Intra-clone",
        bins=bins,
        stat="probability",
        color=color_data,
        alpha=0.5,
    )
    # ax.legend()
    plt.legend(loc=[1.05, 0.4])
    # ax.set_xlabel("Sister-cell distance")
    plt.xlabel("Average similarity score")
    plt.ylabel("Normalized frequency")
    if pvalue_all > 0.0001:
        plt.title(r" $p=$" + f"{pvalue_all:.4f}")
    else:
        # plt.title(f'{source_map[source]}: Pvalue={pvalue:.2E}')
        plt.title(r" $p=$" + f"{pvalue_all:.2E}")
    plt.tight_layout()
    plt.savefig(f"{fig_dir}/{save_data_des}_memory_pvalue.pdf")

    return background, mean_similarity_list, pvalue_all


def embedding_analysis(
    adata,
    clone_key,
    df_sample=None,
    UMAP_parameters=[10, 10, 0.5],
    normalization=True,
    **kwargs,
):
    if df_sample is not None:
        metadata.update_sample_info_on_adata(adata, df_sample)

    adata.obs[clone_key] = adata.obs[clone_key].astype(str)
    df_tmp = pd.DataFrame(adata.obs[clone_key]).dropna()
    clone_names = df_tmp[clone_key].to_numpy()
    cell_id_names = df_tmp.index
    cs.pp.get_X_clone(adata, cell_id_names, clone_names)
    cs.pp.initialize_adata_object(adata)

    X_adjacency = similarity.rescale_similarity_matrix(adata.obsm["X_similarity"])
    if normalization:
        print("spectrum embedding (with normalization)")
        from sklearn.manifold import spectral_embedding

        eig_map = spectral_embedding(
            X_adjacency,
            n_components=50,
        )
        adata.obsm["X_eig"] = eig_map
        if UMAP_parameters[1] >= eig_map.shape[1]:
            UMAP_parameters[1] = eig_map.shape[1]

    else:
        eig, vec = np.linalg.eig(X_adjacency)
        adata.obsm["X_eig"] = np.real(vec[:, 0:50])

    sc.pp.neighbors(
        adata, n_neighbors=UMAP_parameters[0], n_pcs=UMAP_parameters[1], use_rep="X_eig"
    )  # use n_pcs=50 may be better
    sc.tl.umap(adata, min_dist=UMAP_parameters[2])
    adata.obsm["X_emb"] = adata.obsm["X_umap"]

    plotting.plot_multiple_clones_on_embedding(adata, clone_key, **kwargs)
    return adata



import os

import numpy as np
import ot
import pandas as pd
import scipy.stats as stats
import seaborn as sns
from matplotlib import pyplot as plt

from . import lineage
from .settings import *


def lineage_accuracy_from_leaf_order(ordered_clone_array, remove_nan=True):
    """
    The input should be like this:
    ['nan', 'nan', 'nan', '37', 'nan', ...,'nan', 'nan', '37', '2', '2', '2']
    """

    ##remove_nan
    ordered_clone_array = np.array(ordered_clone_array)
    nonnan_idx = np.array(["nan" not in x for x in ordered_clone_array])
    if remove_nan:
        ordered_clone_array_new = ordered_clone_array[nonnan_idx]
    else:
        ordered_clone_array_new = ordered_clone_array

    if len(ordered_clone_array_new) == 0:
        print("no valid clone ids to compute accuracy")
        return pd.DataFrame(
            {
                "clone": ["nan"],
                "clone_size": ["nan"],
                "accuracy": ["nan"],
                "continuity": ["nan"],
                "entropy": ["nan"],
                "wassertein": ["nan"],
            }
        )

    ## compute df_accuracy
    clone_set = list(set(ordered_clone_array_new))
    clone_accuracy = []
    clone_size = []
    clone_max = []
    for x in clone_set:
        tot_N = np.sum(ordered_clone_array_new == x)
        clone_size.append(tot_N)
        continuous_sum = 0
        max_sum = 0
        for j in range(len(ordered_clone_array_new)):
            if ordered_clone_array_new[j] == x:
                continuous_sum = continuous_sum + 1
                if continuous_sum > max_sum:
                    max_sum = continuous_sum
            else:
                continuous_sum = 0
        if max_sum > 1:
            clone_accuracy.append(max_sum / tot_N)
        else:
            clone_accuracy.append(0)
        clone_max.append(max_sum)
    df_accuracy = pd.DataFrame(
        {"clone": clone_set, "clone_size": clone_size, "accuracy": clone_accuracy}
    )
    df_size = pd.DataFrame(
        {"clone": clone_set, "clone_size": clone_size, "clone_max": clone_max}
    )

    #### compute df_entropy
    counter = 0
    x0 = ordered_clone_array_new[0]
    class_list = [0]
    for x in ordered_clone_array_new[1:]:
        if x == x0:
            class_list.append(counter)
        else:
            counter += 1
            x0 = x
            class_list.append(counter)
    df = pd.DataFrame({"clone": ordered_clone_array_new, "class": class_list})
    df_1 = (
        df.groupby(["clone", "class"])
        .apply(lambda x: len(x))
        .reset_index()
        .rename(columns={0: "count"})
    )
    entropy_dict = {}
    for j, x in enumerate(df_1["clone"].unique()):
        df_sub = df_1[df_1["clone"] == x]
        size = df_sub["count"].sum()
        df_sub["prob"] = df_sub["count"] / size
        prob = df_sub["prob"].to_numpy()
        entropy_dict[x] = (np.sum(-prob * np.log(prob))) / np.log(size)
    df_entropy = pd.DataFrame(
        {"clone": entropy_dict.keys(), "entropy": entropy_dict.values()}
    )

    ### compute df_block
    df_block = df_1[["clone"]]
    df_block["block_count"] = df_block.groupby("clone")["clone"].transform("count")
    df_block = df_block.drop_duplicates("clone")
    df_block = df_block.merge(df_size, on="clone")
    df_block["continuity"] = 1 - (df_block["block_count"] - 1) / df_block["clone_size"]
    # print(df_block)
    df_block = df_block[["clone", "continuity"]]

    ### compute df_wassertein
    clone_vector_list = []
    reference_vector_list = []
    Wassertein_list = []
    for clone in clone_set:
        total_clone_length = len(ordered_clone_array_new)
        clone_vector_list = (
            (ordered_clone_array_new == clone).astype(int)
        ).tolist()  # constructing a list of resulting clones
        clone_length = sum(clone_vector_list)  # need to be clone specific

        cover_num = 0
        for p in range(total_clone_length - int(clone_length) + 1):
            new_cover_num = sum(clone_vector_list[p : p + int(clone_length)])
            if cover_num < new_cover_num:
                cover_num = new_cover_num
                ref_start = p
        # constructing a list of reference clones
        reference_vector_list = [
            1 if ref_start <= n < ref_start + int(clone_length) else 0
            for n in range(total_clone_length)
        ]
        input_t0 = np.array(clone_vector_list).astype(float)
        input_t1 = np.array(reference_vector_list).astype(float)

        M = np.where(np.eye(len(input_t0)), 0, 1).astype(int)
        Wd = ot.emd2(
            input_t0, input_t1, M
        )  # compute the Wassterin distance between the observed clone vector and the reference vector
        Wassertein_list.append(Wd / clone_length)
    df_wass = pd.DataFrame({"clone": clone_set, "wassertein": Wassertein_list})

    ### merge results
    df_result = (
        df_accuracy.merge(df_block, on="clone")
        .merge(df_entropy, on="clone")
        .merge(df_wass, on="clone")
        .sort_values(by="clone")
    )
    return df_result[df_result["clone_size"] > 1].filter(
        ["clone", "clone_size", "accuracy", "continuity", "entropy", "wassertein"]
    )


def compute_accuracy_by_depth(
    my_tree, orig_clone_order, method="mean", accuracy_key="continuity"
):
    curr_list = [my_tree]
    depth_count = 0
    accuracy_list = []

    while len(curr_list) > 0:
        next_list = []
        for t in curr_list:
            order_x = np.array(t.get_leaf_names()).astype(int)
            ordered_clone_array = np.array(orig_clone_order).astype(str)[order_x]
            if list(set(ordered_clone_array)) != ["nan"]:
                df_accuracy = lineage_accuracy_from_leaf_order(ordered_clone_array)
                # df_accuracy = df_accuracy[df_accuracy["clone_size"] > 1]
                if len(df_accuracy) > 0:
                    # display(df_1)
                    if method == "mean":
                        mean_accu = df_accuracy[accuracy_key].mean()
                    else:
                        mean_accu = np.sum(
                            df_accuracy[accuracy_key]
                            * df_accuracy["clone_size"]
                            / df_accuracy["clone_size"].sum()
                        )
                    accuracy_list.append(
                        [
                            depth_count,
                            "@".join(df_accuracy["clone"]),
                            df_accuracy["clone_size"].sum(),
                            mean_accu,
                        ]
                    )

            if not t.is_leaf():
                next_list += t.children

        depth_count += 1
        curr_list = next_list
    df = pd.DataFrame(
        accuracy_list, columns=["Depth", "clone", "total_cell_N", accuracy_key]
    )
    return df


def get_accuracy_report(
    out_dir, groupby="lineage", clone_key=None, accuracy_key="accuracy"
):
    """
    groupby can be: 'lineage','patient'
    """

    if clone_key is not None:
        groupby = clone_key + "_" + groupby

    data_des = f"accuracy_{groupby}_"
    all_files = [x for x in os.listdir(out_dir) if x.startswith(data_des)]
    mean_accuracy = []
    weighted_accuracy = []
    tot_clones = []
    for x in all_files:
        df = pd.read_csv(f"{out_dir}/{x}")
        df = df[df["clone_size"] > 1]
        tot_clones.append(len(df))
        weighted_accuracy_tmp = np.sum(
            df[accuracy_key] * df["clone_size"] / df["clone_size"].sum()
        )
        mean_accuracy_tmp = df[accuracy_key].mean()

        mean_accuracy.append(mean_accuracy_tmp)
        weighted_accuracy.append(weighted_accuracy_tmp)

    df_accuracy = pd.DataFrame(
        {
            "FileName": all_files,
            "Total_clone": tot_clones,
            f"mean_{accuracy_key}": mean_accuracy,
            f"weighted_{accuracy_key}": weighted_accuracy,
        }
    )
    df_accuracy["Metadata"] = df_accuracy["FileName"].apply(
        lambda x: x.split(data_des)[1].split("_readcutoff3_minCov")[0]
    )
    for z in ["readcutoff", "minCov", "ExcludeSex"]:
        df_accuracy[z] = df_accuracy["FileName"].apply(
            lambda x: x.split(z)[1].split("_")[0] if z in x else None
        )
    df_accuracy["Stage"] = df_accuracy["Metadata"].apply(lambda x: x.split("_")[0])
    df_accuracy["Annotation"] = df_accuracy["Metadata"].apply(
        lambda x: "_".join(x.split("_")[1:])
    )
    df_accuracy["Source"] = df_accuracy["FileName"].apply(
        lambda x: x.split("_minCov")[1].split("_")[1]
    )

    sel_columns = [
        "Stage",
        "Source",
        "Annotation",
        "readcutoff",
        "minCov",
        "ExcludeSex",
        "Total_clone",
        f"mean_{accuracy_key}",
        f"weighted_{accuracy_key}",
    ]
    df_accuracy = df_accuracy.filter(sel_columns).sort_values(["Stage", "Annotation"])
    for z in ["Source", "ExcludeSex"]:
        df_accuracy[z] = df_accuracy[z].apply(
            lambda x: x.split(".csv")[0] if ((x is not None) and (".csv" in x)) else x
        )
    return df_accuracy


def accuracy_at_different_similarity_threshold(
    X_similarity, my_tree, clone_array, accuracy_key="continuity"
):
    clone_array = np.array(clone_array).astype(str)
    unique_clone_set = list(set(clone_array))
    for x0 in unique_clone_set:
        if np.sum(clone_array == x0) == 1:
            clone_array[clone_array == x0] = "nan"

    ################## Actually identify clones
    lineage.traverse_tree_to_update_dist(my_tree, X_similarity, method="mean")
    S_array = []
    for tmp_tree in my_tree.traverse():
        S_array.append(tmp_tree.dist)
    S_array = np.array(S_array)
    S_array = S_array[~np.isnan(S_array)]

    plt.subplots(figsize=(4, 3.5))
    sns.histplot(S_array, bins=20)
    plt.xlabel("Similarity at branches")
    plt.ylabel("Count")

    mean_accuracy = []
    mean_entropy = []
    relative_cell_fraction = []
    used_threshold_list = []
    step = (np.max(S_array) - np.min(S_array)) / 20
    weight_threshold_list = np.arange(np.min(S_array), np.max(S_array), step)
    for weight_threshold in weight_threshold_list:

        def traverse_tree(node, weight):
            if node.dist >= weight:
                # print(node)
                leaf_idx = np.array(node.get_leaf_names()).astype(int)
                unique_clone_set_tmp = list(set(clone_array[leaf_idx]))
                if unique_clone_set_tmp != ["nan"]:
                    df_tmp = pd.DataFrame({"cell_id": leaf_idx})
                    df_tmp["predicted_clone"] = f"clone_{len(df_list)}"
                    df_tmp["clone_weight"] = node.dist
                    df_tmp["clone_size"] = len(leaf_idx)

                    df_list.append(df_tmp)

            else:
                for child in node.children:
                    traverse_tree(child, weight)

        df_list = []
        traverse_tree(my_tree, weight_threshold)
        if len(df_list) > 0:
            df_result = pd.concat(df_list)
            cell_idx = df_result["cell_id"].to_numpy()
            df_accuracy = lineage_accuracy_from_leaf_order(clone_array[cell_idx])
            mean_accuracy.append(df_accuracy[accuracy_key].mean())
            relative_cell_fraction.append(len(df_result) / len(clone_array))
            used_threshold_list.append(weight_threshold)
    df_final = pd.DataFrame(
        {
            "weight_threshold": used_threshold_list,
            accuracy_key: mean_accuracy,
            "cell_fraction": relative_cell_fraction,
        }
    )
    df_final = (
        df_final.groupby("cell_fraction")
        .agg(
            {
                "weight_threshold": "mean",
                accuracy_key: "mean",
            }
        )
        .reset_index()
        .sort_values("cell_fraction")
    )

    # compute AUC score
    cell_fraction_array = [0] + list(df_final["cell_fraction"])
    accuracy_list = list(df_final[accuracy_key])
    accuracy_array = np.array([accuracy_list[0]] + list(accuracy_list))
    accuracy_array = (accuracy_array[:-1] + accuracy_array[1:]) / 2
    AUC = np.sum(accuracy_array * np.diff(cell_fraction_array))

    # plot
    plt.subplots(figsize=(4, 3.5))
    plt.plot(df_final["weight_threshold"], df_final[accuracy_key], "-*r")
    plt.title(f"{AUC:.4f}")
    plt.xlabel("Similarity threshold")
    plt.ylabel(f"Mean {accuracy_key}")
    plt.xlim([0, 1.02])
    plt.ylim([0, 1.02])
    return df_final


def accuracy_after_removing_leaves_with_low_support(
    df_predict, adata, clone_key, accuracy_key="continuity"
):
    """
    accuracy after filtering out unreliable cells with low tree support
    """
    df_predict["cell_id_new"] = np.array(
        df_predict["cell_id"].apply(lambda x: x.split("-")[0])
    ).astype(int)
    df_predict["predict_large_clone"] = df_predict["clone_size"] > 1
    adata.obs["predicted_large_clone"] = df_predict.sort_values("cell_id_new")[
        "predict_large_clone"
    ].to_numpy()
    sel_idx = ~adata.obs[
        "predicted_large_clone"
    ]  # pd.isna(adata.obs["inferred_clones"])
    old_clone_array = adata.obs[clone_key].copy().to_numpy()
    old_clone_array[sel_idx] = np.nan
    # switch to new clone_key
    ordered_clone_array = old_clone_array[adata.uns["order_x"]].astype(str)
    df_accuracy = lineage_accuracy_from_leaf_order(ordered_clone_array)
    mean_accuracy = df_accuracy[accuracy_key].mean()
    print(
        f"Remaining cell fraction: {np.mean(~sel_idx):.2f};\nClone_N: {len(df_accuracy)};\nMean {accuracy_key}: {mean_accuracy:.2f}; "
    )


def methy_clone_diversity_index(df_predict):
    df_predict["barcode"] = df_predict["cell_id"].apply(lambda x: x.split("-")[-1])

    def observed_clone_size(x):
        x = np.array(x)
        return len(x[x != "nan"])

    def concatenate_clone(x):
        x = list(set(x))
        if "nan" in x:
            x.remove("nan")
        return "_".join(x)

    def barcode_count(x):
        x = list(set(x))
        if "nan" in x:
            x.remove("nan")
        return len(x)

    df_clone_info = df_predict.groupby(["predicted_clone"]).agg(
        barcode_clone_size=("barcode", observed_clone_size),
        barcode_list=("barcode", concatenate_clone),
        barcode_count=("barcode", barcode_count),
    )
    valid_methy_clones = len(
        df_clone_info[
            (df_clone_info["barcode_clone_size"] > 1)
            & (df_clone_info["barcode_count"] == 1)
        ]["barcode_list"].unique()
    )
    merged_methy_clones = np.sum(df_clone_info["barcode_count"] > 1)
    # print(f"Methy clone diversity score: {methy_clone_diversity}")
    return df_clone_info, valid_methy_clones, merged_methy_clones


#####################################

# correlation between met and rna etc

#####################################


def compute_correlation(
    df_data,
    key_1="acc",
    key_2="met",
    corr_by="id",
    cutoff_met_N=2,
    cutoff_acc_N=2,
    mini_size=20,
    save=None,
):
    key_map = {"acc": "acc_rate", "met": "met_rate", "rna": "rna_exp"}
    actual_key_1 = key_map[key_1]
    actual_key_2 = key_map[key_2]
    keep_keys = [
        actual_key_1,
        actual_key_2,
        f"pass_{key_1}QC",
        f"pass_{key_2}QC",
        corr_by,
    ]
    if "met" in [key_1, key_2]:
        keep_keys.append("met_N")
    if "acc" in [key_1, key_2]:
        keep_keys.append("acc_N")

    df_sub = df_data.filter(keep_keys).dropna()
    custom_filter = df_sub[f"pass_{key_1}QC"] & df_sub[f"pass_{key_1}QC"]
    if "met" in [key_1, key_2]:
        keep_keys.append("met_N")
        custom_filter = custom_filter & (df_sub["met_N"] > cutoff_met_N)
    if "acc" in [key_1, key_2]:
        keep_keys.append("acc_N")
        custom_filter = custom_filter & (df_sub["acc_N"] > cutoff_acc_N)

    # series=df_sub[custom_filter].groupby(corr_by).apply(lambda x: np.corrcoef(x[actual_key_1],x[actual_key_2])[0,1])
    df_out = (
        df_sub[custom_filter]
        .groupby(corr_by)
        .apply(
            lambda x: (
                stats.pearsonr(x[actual_key_1], x[actual_key_2])
                if len(x) > mini_size
                else (np.nan, np.nan)
            )
        )
    )
    df_final = (
        pd.DataFrame(df_out.to_list(), index=df_out.index)
        .rename(columns={0: "Corr", 1: "Pvalue"})
        .dropna()
        .assign(log10Pvalue=lambda x: -np.log10(x["Pvalue"]))
    )

    fig, axs = plt.subplots(1, 3, figsize=(12, 4))
    ax = sns.histplot(data=df_final, x="Corr", bins=20, ax=axs[0])
    ax.set_xlabel(f"Corr: {key_1}-{key_2}")
    ax.set_ylabel(f"Histogram")
    ax.set_title(f"Mean: {df_final.Corr.mean():.3f}")
    ax = sns.histplot(data=df_final, x="Corr", y="log10Pvalue", bins=30, ax=axs[1])
    ax.set_xlabel(f"Corr: {key_1}-{key_2}")
    ax.set_ylabel(f"-log10Pvalue")
    ax.set_title(f"Correlation by {corr_by}")
    ax = sns.scatterplot(data=df_final, x="Corr", y="log10Pvalue", ax=axs[2], s=10)
    ax.set_xlabel(f"Corr: {key_1}-{key_2}")
    ax.set_ylabel(f"-log10Pvalue")
    ax.set_title(f"Correlation by {corr_by}")
    fig.tight_layout()
    if save is not None:
        fig.savefig(save)

    return df_final


def compute_coarse_grained_correlation(
    df_data,
    key_1="acc",
    key_2="met",
    corr_by="id",
    coarse_grained_by="celltype",
    mini_size=1,
):
    key_map = {"acc": "acc_rate", "met": "met_rate", "rna": "rna_exp"}
    actual_key_1 = key_map[key_1]
    actual_key_2 = key_map[key_2]
    custom_filter = df_data[f"pass_{key_1}QC"] & df_data[f"pass_{key_1}QC"]
    df_metaccrna_v1 = (
        df_data[custom_filter]
        .groupby(["id", coarse_grained_by, "anno"])
        .agg(
            acc_rate=("acc_rate", "mean"),
            met_rate=("met_rate", "mean"),
            rna_exp=("rna_exp", "mean"),
        )
        .reset_index()
    )

    # coarse-grain by counting all the met sites and unmet sites does not work as well, suggsting strong batch difference across cells
    # df_metaccrna_v1=df_data[custom_filter].groupby(['id',coarse_grained_by,'anno']).agg(met_Nmet=('met_Nmet','sum'),met_N=('met_N','sum'),
    #     acc_Nmet=('acc_Nmet','sum'),acc_N=('acc_N','sum'),rna_exp=('rna_exp','mean')).reset_index().assign(met_rate=lambda df: df['met_Nmet']/df['met_N']).assign(acc_rate=lambda df: df['acc_Nmet']/df['acc_N'])

    df_out = df_metaccrna_v1.groupby("id").apply(
        lambda x: (
            stats.pearsonr(x[actual_key_1], x[actual_key_2])
            if len(x) > 3
            else (np.nan, np.nan)
        )
    )

    df_final = (
        pd.DataFrame(df_out.to_list(), index=df_out.index)
        .rename(columns={0: "Corr", 1: "Pvalue"})
        .dropna()
        .assign(log10Pvalue=lambda x: -np.log10(x["Pvalue"]))
    )

    fig, axs = plt.subplots(1, 3, figsize=(12, 4))
    ax = sns.histplot(data=df_final, x="Corr", bins=20, ax=axs[0])
    ax.set_xlabel(f"Corr: {key_1}-{key_2}")
    ax.set_ylabel(f"Histogram")
    ax.set_title(f"Mean: {df_final.Corr.mean():.3f}")
    ax = sns.histplot(data=df_final, x="Corr", y="log10Pvalue", bins=30, ax=axs[1])
    ax.set_xlabel(f"Corr: {key_1}-{key_2}")
    ax.set_ylabel(f"-log10Pvalue")
    ax.set_title(f"Correlation by {corr_by}")
    ax = sns.scatterplot(data=df_final, x="Corr", y="log10Pvalue", ax=axs[2], s=10)
    ax.set_xlabel(f"Corr: {key_1}-{key_2}")
    ax.set_ylabel(f"-log10Pvalue")
    ax.set_title(f"Correlation by {corr_by}")
    plt.tight_layout()

    return df_final, df_metaccrna_v1



import os
import time

import multiprocess
import numpy as np
import pandas as pd
import scanpy as sc

from . import metadata, scripts
from .settings import *

###########################
# build adata from raw data
###########################


def generate_similarity_heatmap_for_pseudo_cells(
    data_path,
    selected_key,
    annotation,
    similarity_correction=0,
    source="met",
    threads=10,
    selection_list=None,
    reference="mm10",
    min_cells=5,
    exclude_sex=0,
    read_cutoff=3,
    min_coverage=2,
    conda_script="source /home/wangshouwenLab/wangshouwen/miniconda3/etc/profile.d/conda.sh; conda activate CoSpar_test",
):
    df_sample = metadata.load_sample_info(data_path)
    if selection_list is None:
        # selection_list=list(df_sample[selected_key].dropna().unique())
        valid_rows = df_sample[selected_key].notnull()
        value_counts = df_sample[selected_key][valid_rows].value_counts()
        selected_values = value_counts[value_counts > min_cells].index
        selection_list = list(selected_values)

    pool = multiprocess.Pool(threads)

    ## generate methylation profile for each individual pseudo-cell
    def generate_pseudobulk(sample_tmp):
        script_name = scripts.generate_script_bulk_bigwig(
            data_path,
            source=source,
            group_by=selected_key,
            sample_list=[sample_tmp],
            reference=reference,
            scenario="only_bulk",
            min_cells=min_cells,
        )
        command = f"{conda_script}; python {script_name}"
        os.system(command)

    pool.map(generate_pseudobulk, selection_list)

    ## make the corresponding directories
    pseudo_bulk_data = f"{data_path}/met/cpg_level/pseudobulk/{selected_key}"
    data_path_pseudo = f"{data_path}/pseudocell/{selected_key}/downstream_R/all_data"
    os.makedirs(f"{data_path_pseudo}/met/cpg_level", exist_ok=True)
    os.system(f"ln -sf {pseudo_bulk_data}/* {data_path_pseudo}/met/cpg_level")

    scripts.generate_cell_by_region_matrix(
        data_path_pseudo, selection_list, annotation, threads=threads
    )

    df_sample_coarse = (
        df_sample.groupby(selected_key)
        .agg({"nCG": "sum", "met_rate": "mean"})
        .reset_index()
        .rename(columns={selected_key: "sample"})
    )
    df_sample_coarse["HQ"] = True
    df_sample_coarse[selected_key] = df_sample_coarse["sample"]
    # df_sample_coarse['cell_type']=df_sample_coarse[selected_key].apply(lambda x: x.split('_')[0])
    df_sample_coarse.to_csv(
        f"{data_path_pseudo}/sample_sheet.tsv.gz", compression="gzip", sep="\t", index=0
    )

    groupby = selected_key
    selected = "all"
    source = "met"
    out_dir = f"{data_path_pseudo}/out_dir"
    os.makedirs(out_dir, exist_ok=True)
    clone_key = selected_key  #'cell_type'
    script_name = f"{source_script_dir}/DNA/lineage_reconstruction_from_DNA_met.py"
    save_data_des = f"{groupby}_{selected}_{annotation}_readcutoff{read_cutoff}_minCov{min_coverage}_{source}_ExcludeSex{exclude_sex}_PseudoCell"
    command = f"""
             {conda_script}; python {script_name} \
               --root_dir {data_path_pseudo}  --out_dir {out_dir}  --annotation {annotation} --groupby {groupby} --selected {selected}  --read_cutoff {read_cutoff} \
                --min_cov {min_coverage}  --clone_key {clone_key}  --source {source}  --exclude_sex {exclude_sex}  --save_data_des {save_data_des}   --compute_accuracy 0 \
                --similarity_correction {similarity_correction}
            """
    os.system(command)
    adata_path = f"{out_dir}/{save_data_des}.h5ad"
    print("adata path:")
    print(adata_path)
    return sc.read(adata_path)


def generate_similarity_heatmap_for_pseudo_cells_from_more_info(
    data_path,
    selected_key,
    annotation,
    additional_key=None,
    similarity_correction=0,
    source="met",
    threads=10,
    selection_list=None,
    reference="mm10",
    min_cells=5,
    conda_script="source /home/wangshouwenLab/wangshouwen/miniconda3/etc/profile.d/conda.sh; conda activate CoSpar_test",
):

    df_sample = metadata.load_sample_info(data_path)
    selected_key_list = []
    adata_list = []
    for x in df_sample[additional_key].dropna().unique():
        selected_key_list.append(f"{x}_{selected_key}")
        if f"{x}_{selected_key}" not in df_sample.columns:
            mask = df_sample[additional_key] == x
            df_sample.loc[mask, f"{x}_{selected_key}"] = (
                df_sample[mask][additional_key] + "_" + df_sample[mask][selected_key]
            )
    metadata.backup_and_save_sample_info(df_sample, data_path)

    for selected_key in selected_key_list:
        selection_list = list(df_sample[selected_key].dropna().unique())
        adata_tmp = generate_similarity_heatmap_for_pseudo_cells(
            data_path,
            selected_key,
            annotation,
            source=source,
            threads=threads,
            selection_list=selection_list,
            similarity_correction=similarity_correction,
            reference=reference,
            min_cells=min_cells,
            conda_script=conda_script,
        )
        adata_list.append(adata_tmp)
    return adata_list


def generate_single_CpG_adata(
    cpg_level_path,
    valid_sample_list,
    read_cutoff=1,
    merge_strands=False,
    CpG_ref="mm10",
    selected_region_singleCpG=None,
    exclude_sex=0,
    cores=64,
):

    def load_sample_df(sample):
        df_data_tmp = pd.read_csv(f"{cpg_level_path}/{sample}.tsv.gz", sep="\t")
        df_data_tmp.columns = ["chr", "pos", "met_reads", "nonmet_reads", "rate"]
        df_data_tmp["sample"] = sample
        return df_data_tmp

    if cores > len(valid_sample_list):
        cores = len(valid_sample_list)

    from joblib import Parallel, delayed

    df_list = Parallel(n_jobs=cores)(
        delayed(load_sample_df)(i) for i in valid_sample_list
    )

    df_all = pd.concat(df_list, ignore_index=True)

    # Clear the list to remove references
    df_list.clear()

    print(f"filter chr; time {time.time()}")
    # exclude chrM, and also other regions
    df_all["chr"] = df_all["chr"].astype(str)
    chr_dict = {str(x): f"chr{x}" for x in range(1, 30)}
    chr_dict.update({f"chr{x}": f"chr{x}" for x in range(1, 30)})
    chr_dict.update({"X": "chrX", "Y": "chrY", "chrX": "chrX", "chrY": "chrY"})
    df_all["chr"] = df_all["chr"].map(chr_dict)
    df_all = df_all[~pd.isna(df_all["chr"])]

    print(f"Add chr_pos; time {time.time()}")
    chr_pos = [
        f"{x}_{y}" for x, y in zip(df_all["chr"].to_list(), df_all["pos"].to_list())
    ]
    df_all["chr_pos"] = chr_pos

    if selected_region_singleCpG is not None:
        sel_reg_singleCpG = df_all["chr_pos"].isin(selected_region_singleCpG)
        df_all = df_all[sel_reg_singleCpG]

    if merge_strands:
        if CpG_ref not in ["hg38", "hg19", "mm10", "mm9"]:
            raise ValueError(f"CpG_ref {CpG_ref} is not found")
        CpG_ref_dir = f"{CpG_ref_location}/{CpG_ref}.CpGs_new.bed"
        ##############################
        ## Merge CpGs from two strands: BEGIN
        ###############################
        print(f"Merge CpGs from two strands; time {time.time()}")
        df_ref = pd.read_csv(
            CpG_ref_dir,
            sep="\t",
            names=["chr", "pos", "name", "chr_pos", "chr_pos_neg"],
        )
        strand_dict = dict(
            zip(df_ref["chr_pos_neg"].to_list(), df_ref["chr_pos"].to_list())
        )
        sel_idx = df_all["chr_pos"].isin(df_ref["chr_pos"].to_list())
        df_all_pos = df_all[sel_idx]  # CpGs from the positive strand
        df_all_neg = df_all[~sel_idx]  # CpGs from the negative strand
        df_all_neg["chr_pos"] = df_all_neg["chr_pos"].map(strand_dict).to_list()
        nan_frac = np.sum(pd.isna(df_all_neg["chr_pos"])) / len(df_all_neg)
        print(f"Fraction without mapping: {nan_frac}; time {time.time()}")
        if nan_frac > 0.35:
            print("****************ERROR************************")
            print(
                f"{nan_frac} not mapped. Likely due to use the wrong reference genome. Current reference is {CpG_ref}"
            )
            print("****************ERROR************************")
        df_all_neg_clean = df_all_neg[~pd.isna(df_all_neg["chr_pos"])]
        df_all = pd.concat([df_all_pos, df_all_neg_clean])
        del df_all_pos, df_all_neg, df_all_neg_clean, df_ref, strand_dict
        tot_N = len(df_all)
        df_all = (
            df_all.groupby(["chr_pos", "sample"])
            .agg({"met_reads": "sum", "nonmet_reads": "sum"})
            .reset_index()
        )
        drop_frac = 1 - len(df_all) / tot_N
        print(f"Fraction of reads that are dropped: {drop_frac:.2f}")
        if drop_frac > 0.35:
            print("****************ERROR************************")
            print(
                f"{drop_frac} dropped. Likely due to use the wrong reference genome. Current reference is {CpG_ref}"
            )
            print("****************ERROR************************")

        ##############################
        ## Merge CpGs from two strands: END
        ###############################

    print(f"total reads; time {time.time()}")
    df_all["total_reads"] = (
        df_all["met_reads"].to_numpy() + df_all["nonmet_reads"].to_numpy()
    )
    print(f"filtering by reads and rate; time {time.time()}")
    if read_cutoff > 1:
        df_all = df_all[df_all["total_reads"] >= read_cutoff]

    df_all["rate"] = df_all["met_reads"] / df_all["total_reads"]

    cutoff = 0.01
    df_all = df_all[~((df_all["rate"] > cutoff) & (df_all["rate"] < (1 - cutoff)))]
    df_all.loc[df_all["rate"] <= cutoff, "rate"] = 0
    df_all.loc[df_all["rate"] >= (1 - cutoff), "rate"] = 1

    print(f"pivot and sub_adata generation; time {time.time()}")
    adata = sc.AnnData(
        pd.pivot(df_all, index="sample", columns="chr_pos", values="rate")
    )
    del df_all

    if exclude_sex > 0:
        print(f"Exclude sex chromosome; time {time.time()}")
        var_name_array = np.array([x.split("_")[0] for x in adata.var_names])
        full_var_idx = (var_name_array != "chrX") & (var_name_array != "chrY")
        adata = adata[:, full_var_idx]
    elif exclude_sex < 0:
        print(f"Use only sex chromosome; time {time.time()}")
        var_name_array = np.array([x.split("_")[0] for x in adata.var_names])
        full_var_idx = (var_name_array == "chrX") | (var_name_array == "chrY")
        adata = adata[:, full_var_idx]

    return adata


####################
# Methscan related
####################


def generate_adata_from_methscan_matrix(df_data, matrix_dir, read_cutoff=1):
    df_matrix = df_data[(df_data["coverage"] >= read_cutoff)].pivot(
        columns="col", values="mfracs", index="row"
    )
    df_barcode = pd.read_csv(
        f"{matrix_dir}/barcodes.tsv.gz", compression="gzip", header=None
    )
    df_feature = pd.read_csv(
        f"{matrix_dir}/features.tsv.gz", compression="gzip", header=None
    )

    adata = sc.AnnData(df_matrix.to_numpy())
    adata.obs_names = np.array(df_barcode[0].to_list())[
        np.array(df_matrix.index.to_list()) - 1
    ]
    adata.var_names = np.array(df_feature[0].to_list())[
        np.array(df_matrix.columns.to_list()) - 1
    ]
    return adata


def filter_adata(adata, exclude_sex=0, exclude_chrM=True, min_cell_coverage=2):
    var_name_array = np.array([x.split(":")[0] for x in adata.var_names])
    if exclude_chrM:
        full_var_idx = var_name_array != "chrM"
        adata = adata[:, full_var_idx]  # remove Mitochontria

    ## removing sex related features if needed
    if exclude_sex > 0:
        print("Exclude sex chromosome")
        full_var_idx = (var_name_array != "chrX") & (var_name_array != "chrY")
        adata = adata[:, full_var_idx]
    elif exclude_sex < 0:
        print(f"Use only sex chromosome")
        full_var_idx = (var_name_array == "chrX") | (var_name_array == "chrY")
        adata = adata[:, full_var_idx]

    print(f"full adata shape: {adata.shape}")

    ## filter by coverage
    sel_index = (~np.isnan(adata.X)).sum(
        0
    ) >= min_cell_coverage  # min_coverage * adata.shape[0]
    adata = adata[:, sel_index]
    print(
        f"adata shape removing regions covering less than {min_cell_coverage} cells : {adata.shape}"
    )
    return adata


def load_methscan_matrix(matrix_dir):
    df_data = pd.read_csv(
        f"{matrix_dir}/matrix.mtx.gz", compression="gzip", header=None, sep=" "
    )
    df_data.columns = ["row", "col", "residuals", "mfracs", "coverage"]
    return df_data
    return df_data



import os

import cospar as cs
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import rcParams
from mpl_toolkits.axes_grid1 import make_axes_locatable

from .settings import color_list, color_list_2

###########

# color

###########


def plot_color_bar(fig, cax_list, hex_color_list):
    import matplotlib as mpl

    if type(cax_list) != list:
        cax_list = [cax_list]
        hex_color_list = [hex_color_list]

    for j, cax in enumerate(cax_list):
        cmaplist = [hex_to_rgb(__[1:]) for __ in hex_color_list[j]]

        # create the new map
        cmap = mpl.colors.LinearSegmentedColormap.from_list(
            "Custom cmap", cmaplist, len(hex_color_list[j])
        )

        # define the bins and normalize
        bounds = np.linspace(0, len(hex_color_list[j]), len(hex_color_list[j]) + 1)
        norm = mpl.colors.BoundaryNorm(bounds, cmap.N)
        # ax2 = fig.add_axes([0.95, 0.15, 0.03, 0.75])
        cb = mpl.colorbar.ColorbarBase(
            cax,
            cmap=cmap,
            norm=norm,
            spacing="proportional",
            ticks=None,
            boundaries=bounds,
            format="%1i",
        )
        cb.set_ticks([])


def hex_to_rgb(hex):
    rgb = []
    for i in (0, 2, 4):
        decimal = int(hex[i : i + 2], 16)
        rgb.append(decimal)
    rgb.append(256)
    rgb = np.array(rgb) / 256
    return tuple(rgb)


def get_clone_color_dictionary(
    clone_array,
    initial_clone_color_dict=None,
    order_by_size=True,
    color_choice=0,
):
    background_color = "#ffffff"  ##f0f0f0

    if len(set(clone_array)) == len(clone_array):
        print("use pre-set clone order")
        unique_clones = list(clone_array)  # use the pre-set order
    else:
        df = pd.DataFrame({"clone_id": clone_array}).reset_index()
        if order_by_size:
            df_clone = (
                df.groupby("clone_id")
                .agg(cell_count=("index", "count"))
                .sort_values("cell_count", ascending=False)
                .reset_index()
            )

            unique_clones = df_clone["clone_id"].to_list()
        else:
            unique_clones = list(df["clone_id"].unique())

    if (color_choice % 2) == 0:
        color_coding_temp = color_list.copy()
    else:
        color_coding_temp = color_list_2.copy()

    if initial_clone_color_dict is None:
        clone_color_dict = {}
    else:
        clone_color_dict = initial_clone_color_dict
        for clone_tmp, color_tmp in initial_clone_color_dict.items():
            if clone_tmp in unique_clones:
                unique_clones.remove(clone_tmp)
                if color_tmp in color_coding_temp:
                    color_coding_temp.remove(color_tmp)

    if "nan" in unique_clones:
        unique_clones.remove("nan")
    clone_color_dict["nan"] = background_color

    jmax = len(color_coding_temp) - 1
    for j, x in enumerate(unique_clones):
        if j < jmax:
            clone_color_dict[x] = color_coding_temp[j]
        else:
            clone_color_dict[x] = color_coding_temp[jmax]

    return clone_color_dict


##############

# similarity

##############


def plot_similarity_heatmap_with_clone_label(
    input_X,
    label=None,
    label_x=None,
    label_y=None,
    vmax=None,
    vmin=None,
    figsize=(10, 10),
    fontsize=4,
    save_name=None,
    title=None,
    color_bar=False,
    clone_color_dict=None,
    show_label=True,
    cell_index=None,
    legend_labels=None,
    legend_fontsize=10,
    legend_position="upper left",
    show_legend=True,
):
    """
    Since we need to add a color bar, based on the input labels, we cannot change of the order of index within the cs.pl.heatmap function
    """
    cs.settings.set_figure_params(fontsize=fontsize)
    fig, ax = plt.subplots(figsize=figsize)

    if label_x is None:
        label_x = label
    if label_y is None:
        label_y = label

    if show_label:
        if cell_index is not None:
            label_x_tmp = [f"{cell_index[j]}: {x}" for j, x in enumerate(label_x)]
            label_y_tmp = [f"{cell_index[j]}: {x}" for j, x in enumerate(label_y)]
        else:
            label_x_tmp = label_x.copy()
            label_y_tmp = label_y.copy()
    else:
        label_x_tmp = None
        label_y_tmp = None

    cs.pl.heatmap(
        input_X,
        y_ticks=label_y_tmp,
        x_ticks=label_x_tmp,
        order_map_x=False,
        order_map_y=False,
        vmax=vmax,
        vmin=vmin,
        color_bar=color_bar,
        ax=ax,
    )
    if title is not None:
        plt.title(title)

    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="3%", pad=0.1)

    clone_color_dict = get_clone_color_dictionary(
        label_y, initial_clone_color_dict=clone_color_dict
    )
    color_array = [clone_color_dict[__] for __ in label_y]

    plot_color_bar(
        fig, cax, color_array[::-1]
    )  # we flip the order because the cs.pl.heatmap will also flip y axis

    if show_legend:  # Add condition to show legend
        if legend_labels is None:
            legend_labels = list(clone_color_dict.keys())

        color_label = [clone_color_dict[label] for label in legend_labels]
        handles = []
        for label, color in zip(legend_labels, color_label):
            rect = mpatches.Rectangle((0, 0), 1, 1, color=color, label=label)
            handles.append(rect)

        legend = plt.legend(
            handles=handles,
            fontsize=legend_fontsize,
            bbox_to_anchor=(1.1, 1),
            loc=legend_position,
        )
        legend.set_title(None)

    plt.tight_layout()
    if save_name is not None:
        # os.makedirs(save_dir, exist_ok=True)
        plt.savefig(save_name)
        print(f"figure saved at: {save_name}")


def plot_similarity_heatmap_with_multiple_colorbars(
    adata,
    additional_key_list=None,
    heatmap_vmin_percentile=50,
    heatmap_vmax_percentile=99.5,
    figsize=(10, 9.5),
    max_clone_N=20,
    min_clone_size=2,
    fig_dir=None,
    save_name=None,
    data_des=None,
    show_label=False,
    show_x_label=False,
    title=None,
    rotate_cells=False,
    clone_color_dict_orig=None,
    fontsize=6,
    add_cell_index=True,
    show_legend=True,
    legend_fontsize=10,
    maximum_labels_per_column=40,
    legend_position="upper left",
):
    order_x = adata.uns["order_x"]
    ordered_clone_array = adata.uns["ordered_clone_array"]
    ordered_clone_array = exclude_small_clones(
        ordered_clone_array, max_clone_N=max_clone_N, min_clone_size=min_clone_size
    )

    if add_cell_index:
        cell_index = order_x
    else:
        cell_index = None

    if rotate_cells:
        x0 = ordered_clone_array[0]
        count_0 = 0
        unique_clone_set = list(set(ordered_clone_array))
        if "nan" in unique_clone_set:
            unique_clone_set.remove("nan")

        if len(unique_clone_set) > 1:
            while (ordered_clone_array[count_0] == x0) or (
                ordered_clone_array[count_0] == "nan"
            ):
                count_0 += 1
            ordered_clone_array = np.hstack(
                [
                    ordered_clone_array[count_0:],
                    np.array(ordered_clone_array[:count_0])[::-1],
                ]
            )
            order_x = np.hstack([order_x[count_0:], np.array(order_x[:count_0])[::-1]])

    input_X = adata.obsm["X_similarity"][order_x][:, order_x]
    label_list = [ordered_clone_array]
    if additional_key_list is not None:
        for tmp_key in additional_key_list:
            label_tmp = adata.obs[tmp_key].astype(str)[order_x]
            label_tmp = exclude_small_clones(
                label_tmp, max_clone_N=max_clone_N, min_clone_size=min_clone_size
            )
            label_list.append(label_tmp)

    heatmap_vmin = np.percentile(input_X.flatten(), heatmap_vmin_percentile)
    heatmap_vmax = np.percentile(np.triu(input_X, k=1), heatmap_vmax_percentile)

    if heatmap_vmin >= heatmap_vmax:
        print("heatmap_vmax<=heatmap_vmin")
        heatmap_vmin = None
        heatmap_vmax = None

    if show_label:
        if cell_index is not None:
            label_y_tmp = [
                f"{cell_index[j]}: {x}" for j, x in enumerate(label_list[-1])
            ]
        else:
            label_y_tmp = label_list[-1].copy()
    else:
        label_y_tmp = None

    if show_x_label:
        x_ticks = label_y_tmp
    else:
        x_ticks = None

    cs.settings.set_figure_params(fontsize=fontsize)
    fig, ax = plt.subplots(figsize=figsize)
    cs.pl.heatmap(
        input_X,
        y_ticks=label_y_tmp,
        x_ticks=x_ticks,
        order_map_x=False,
        order_map_y=False,
        vmax=heatmap_vmax,
        vmin=heatmap_vmin,
        color_bar=False,
        ax=ax,
    )
    if title is not None:
        plt.title(title)

    divider = make_axes_locatable(ax)

    color_array_list = []
    cax_list = []
    handles_list = []
    for j, label in enumerate(label_list):

        if j == 0:
            clone_color_dict = get_clone_color_dictionary(
                label, initial_clone_color_dict=clone_color_dict_orig, color_choice=j
            )
        else:
            if clone_color_dict_orig is not None:
                missing_clones = set(clone_color_dict.keys()) - set(
                    clone_color_dict_orig.keys()
                )
                for missing_clone in missing_clones:
                    clone_color_dict_orig[missing_clone] = clone_color_dict[
                        missing_clone
                    ]
            else:
                clone_color_dict_orig = clone_color_dict

            clone_color_dict = get_clone_color_dictionary(
                label, initial_clone_color_dict=clone_color_dict_orig, color_choice=j
            )

        color_array = [clone_color_dict[__] for __ in label][::-1]
        color_array_list.append(color_array)

        cax = divider.append_axes("right", size="2.5%", pad=0.1)
        cax_list.append(cax)

        if show_legend:  # Add condition to show legend
            legend_labels = list(clone_color_dict.keys())
            if "nan" in legend_labels:
                legend_labels.remove("nan")
            color_label = [clone_color_dict[label] for label in legend_labels]

            handles = []
            for label, color in zip(legend_labels, color_label):
                rect = mpatches.Rectangle((0, 0), 1, 1, color=color, label=label)
                handles.append(rect)
            handles_list.append(handles.copy())

    plot_color_bar(
        fig, cax_list, color_array_list
    )  # we flip the order because the cs.pl.heatmap will also flip y axis

    if show_legend:
        ncol = int(len(legend_labels) / maximum_labels_per_column) + 1
        legend = plt.legend(
            handles=handles_list[-1],
            fontsize=legend_fontsize,
            bbox_to_anchor=[1.1, 1],
            ncol=ncol,
            loc=legend_position,
        )
        legend.set_title(None)

    plt.tight_layout()

    if save_name is not None:
        plt.savefig(save_name)
    elif (fig_dir is not None) and (data_des is not None):
        save_name = f"{fig_dir}/{data_des}_similarity_matrix_multiple_colorbars.pdf"
        plt.savefig(save_name)


def find_vmax(values, q=99):
    return np.percentile(values, q=q)


def plot_all_stats(
    df_all, figure_path, figsize=(3.5, 3.5), color="#fcae91", data_des=""
):
    ylabel_dict = {
        "mean_accuracy": "Lineage accuracy",
        "memory_strength": "Memory score:\n"
        + r"($\overline{intra.}-\overline{inter.}$)/(std. inter.)",
        "log10Pv": "Signif. of clonal memory:\n" + r"$-$log" + r"$_{10}$(P-value)",
    }

    for sel_key in ["mean_accuracy", "memory_strength", "log10Pv"]:
        fig, ax = plt.subplots(figsize=figsize)

        # 示例数据
        x = df_all["stage"]
        y = df_all[sel_key].values

        # 生成x轴的位置
        x_pos = np.arange(len(x))

        # 绘制柱状图
        plt.bar(x_pos, y, width=0.5, color=color)  # ,fill = False)

        # 设置x轴刻度标签
        plt.xticks(x_pos, x)
        plt.xticks(rotation=90)

        # 设置y轴标签
        plt.ylabel(ylabel_dict[sel_key])

        # 在柱状图的顶部标记额外的数字
        if sel_key == "log10Pv":
            clone_numbers = df_all["clone_N"].values  # 要添加的额外数字列表
            for i in range(len(x_pos)):
                plt.text(
                    x_pos[i],
                    y[i] + max(y) * 0.01,
                    clone_numbers[i],
                    ha="center",
                    va="bottom",
                    fontsize=12,
                )
        if sel_key == "mean_accuracy":
            plt.ylim([0, 1])

        # 显示图形
        # plt.show()
        plt.tight_layout()
        plt.savefig(f"{figure_path}/stats_all_{sel_key}_{data_des}.pdf")


def exclude_small_clones(clone_array, max_clone_N=20, min_clone_size=2):
    df = pd.DataFrame({"clone_id": list(clone_array)}).reset_index()
    df_clone = (
        df.groupby("clone_id")
        .agg(cell_count=("index", "count"))
        .sort_values("cell_count", ascending=False)
        .reset_index()
    )
    if len(df_clone) > max_clone_N:
        df_clone = df_clone.iloc[:max_clone_N]
    df_clone = df_clone[df_clone["cell_count"] >= min_clone_size]
    valid_clones = df_clone["clone_id"].unique()
    new_array = []
    for x in clone_array:
        if x not in valid_clones:
            new_array.append("nan")
        else:
            new_array.append(x)
    return new_array


def display_similarity_matrix(out_dir, SAVE_DATA_DES, clone_key=None):
    from IPython.display import Image, display
    from wand.image import Image as WImage

    if clone_key is None:
        file_name = f"{out_dir}/similarity_matrix_{SAVE_DATA_DES}.pdf"
    else:
        file_name = f"{out_dir}/similarity_matrix_{clone_key}_{SAVE_DATA_DES}.pdf"
    if os.path.exists(file_name):
        display(WImage(filename=file_name))
    else:
        print(f"file not found: {file_name}")


################

# embedding

################


def plot_multiple_clones_on_embedding(
    adata,
    clone_key,
    sel_clone_set=None,
    figure_path=".",
    clone_prefix=None,
    title="",
    initial_clone_color_dict=None,
    max_sel_clone=None,
    figsize=(5, 3.5),
    fontsize=13,
    marker_size=100,
    line_width=0.5,
    color_choice=0,
    save_data_des="",
    basis="X_umap",
    background_size=50,
    maximum_labels_per_column=12,
):

    df_clone_0 = (
        pd.DataFrame({clone_key: list(adata.obs[clone_key])})
        .reset_index()
        .groupby(clone_key)
        .agg({"index": "count"})
        .rename(columns={"index": "clone_size"})
    )
    df_clone_1 = (
        df_clone_0[df_clone_0["clone_size"] > 1]
        .sort_values("clone_size", ascending=False)
        .reset_index()
    )

    all_clones = df_clone_1[clone_key].to_numpy().astype(str)

    if "nan" in list(all_clones):
        nan_symbol = "nan"
    else:
        nan_symbol = "NA"

    if sel_clone_set is None:
        if max_sel_clone is None:
            max_sel_clone = len(df_clone_1)
        sel_clone_set = list(all_clones[:max_sel_clone])  #
        if nan_symbol in sel_clone_set:
            sel_clone_set.remove(nan_symbol)

    if clone_prefix is not None:
        new_order = [
            f"{clone_prefix}_{y}"
            for y in sorted(
                [int(x.split(f"{clone_prefix}_")[1]) for x in sel_clone_set]
            )
        ]
    else:
        new_order = sorted(sel_clone_set)

    adata.obs["Clone ID"] = adata.obs[clone_key].map(
        dict(zip(sel_clone_set, sel_clone_set))
    )
    adata.obs["Clone ID"] = adata.obs["Clone ID"].astype(str)  # turn NAN to 'nan'

    clone_color_dict = get_clone_color_dictionary(
        sel_clone_set,
        initial_clone_color_dict=initial_clone_color_dict,
        color_choice=color_choice,
    )

    #########
    ### plot and save the embedding
    #########
    fig, ax = plt.subplots(figsize=figsize)
    cs.settings.set_figure_params(
        pointsize=marker_size, figsize=figsize, fontsize=fontsize
    )
    ax = cs.pl.plot_adata_with_prefered_order(
        adata,
        "Clone ID",
        plot_order=[nan_symbol] + new_order,
        background="nan",
        palette=clone_color_dict,
        s=marker_size,
        linewidth=line_width,
        edgecolor="k",
        background_size=background_size,
        basis=basis,
    )

    # Get the legend handles
    handles, labels = ax.get_legend_handles_labels()
    labels = np.array(labels)
    handles = np.array(handles)
    sel_idx = labels != "nan"

    # # Iterate through the handles and call `set_edgecolor` on each
    # for ha in handles:
    #     ha.set_edgecolor("k")
    #     ha.set_linewidth(0.5)

    # Use `ax.legend` to set the modified handles and labels
    ncol = int(np.sum(sel_idx) / maximum_labels_per_column) + 1
    lgd = ax.legend(
        handles[sel_idx],
        labels[sel_idx],
        loc=[1.05, 0],
        ncol=ncol,
        frameon=False,
    )
    # plt.legend(frameon=False)
    plt.title(title)

    plt.tight_layout()
    plt.savefig(f"{figure_path}/{save_data_des}_all_clone_{clone_key}.pdf")


#########################################

# tree (related, see viewtree package)

#########################################


def plot_tree_with_support(
    tree_path, figsize=(10, 15), leaf_name_map=None, show_support=True
):
    """
    leaf_name_map can be a dictionary or a function to map the old name to new name.

    Try this:
    leaf_name_map=dict(zip(np.arange(adata_final.shape[0]).astype(str), adata_final.obs_names))
    """
    import matplotlib.pyplot as plt
    from Bio import Phylo
    from ete3 import Tree

    with open(f"{tree_path}", "r") as f:
        if show_support:
            my_tree = Tree(f.read(), format=0)
        else:
            my_tree = Tree(f.read(), format=1)
    if leaf_name_map is not None:
        for leaf in my_tree.iter_leaves():
            if type(leaf_name_map) is dict:
                leaf.name = leaf_name_map[leaf.name]  # This is a dictionary
            else:
                leaf.name = leaf_name_map(leaf.name)  # This is a function

    if show_support:
        my_tree = my_tree.write(format=0)
    else:
        my_tree = my_tree.write(format=1)
    with open(f"{tree_path}_tmp.txt", "w") as f:
        f.write(my_tree)

    tree = Phylo.read(f"{tree_path}_tmp.txt", "newick")

    fig, ax = plt.subplots(figsize=figsize)
    plt.ion()
    Phylo.draw(tree, axes=ax)
    ax.axis("off")
    plt.show()


#########################################

# random

#########################################


def boxplot_for_differential_activity(
    df_data, df_sample, genome_id, genome_id_key="id", source="acc", plot_type="boxplot"
):
    df_gene = (
        df_data[df_data[genome_id_key] == genome_id]
        .merge(df_sample.filter(["sample", "celltype", "HQ"]), on="sample")
        .query("HQ==True")
    )  # Spp1, Dppa2,'Tex19.1'
    fig, ax = plt.subplots()
    if plot_type == "boxplot":
        sns.boxplot(data=df_gene, x="celltype", y="rate")
    else:
        sns.violinplot(data=df_gene, x="celltype", y="rate")
    sns.stripplot(data=df_gene, x="celltype", y="rate", size=4, color=".3", linewidth=0)
    plt.xticks(rotation=90)
    plt.xlabel("")
    if source == "acc":
        plt.ylabel("Accessibility rate")
    else:
        plt.ylabel("Methylation rate")
    plt.title(f"{genome_id}; sample N= {len(df_gene)}")


def visualize_raw_rates(
    df_data_list,
    chr_id="2",
    start_N=105130000,
    end_N=105930000,
    color=["r", "b", "k", "g", "yellow", "cyan"],
    pos_key="pos",
    marker=".",
    y_key="rate",
    chr_key="chr",
    merge_track=False,
):
    """
    chr_id: {'1','2','3'...,'X','Y'}

    Note that y_key can be a str, or a list.
    You will need to run the following command to obtain the best effect
    ```jupyter
    %config InlineBackend.figure_format = 'svg' #'retina'         # or 'svg'
    ```
    """
    import seaborn as sns

    sns.set_style("white")
    rcParams["axes.spines.right"] = False
    rcParams["axes.spines.top"] = False
    rcParams["font.size"] = 13
    rcParams["lines.markersize"] = 10

    if (type(y_key) == str) and (type(df_data_list) == list):
        y_key = [y_key for _ in df_data_list]

    if (type(df_data_list) == pd.DataFrame) and (type(y_key) == list):
        df_data_list = [df_data_list for _ in y_key]

    if (type(df_data_list) == pd.DataFrame) and (type(y_key) == str):
        df_data_list = [df_data_list]
        y_key = [y_key]

    if len(color) < len(df_data_list):
        raise ValueError("Please provide more colors if you want more than 4")

    if merge_track:
        fig, ax = plt.subplots(figsize=(30, len(df_data_list)))
    else:
        fig, axs = plt.subplots(
            len(df_data_list), 1, figsize=(30, 3 * len(df_data_list))
        )
    for j, df_test in enumerate(df_data_list):
        df_test[chr_key] = df_test[chr_key].astype(str)
        df_test[pos_key] = df_test[pos_key].astype(int)

        if chr_id not in list(set(df_test[chr_key])):
            raise ValueError("chr_id should be selected from ", set(df_test[chr_key]))

        if start_N > df_test[pos_key].max():
            raise ValueError("start_N should be smaller than ", df_test[chr_key].max())

        if end_N < df_test[pos_key].min():
            raise ValueError("end_N should be larger than ", df_test[chr_key].min())

        tmp = df_test[df_test[chr_key] == chr_id]
        df_test_0 = tmp[(tmp[pos_key] > start_N) & (tmp[pos_key] < end_N)].sort_values(
            pos_key
        )
        if merge_track:
            ax.plot(df_test_0[pos_key], df_test_0[y_key[j]], marker, color=color[j])
        else:
            axs[j].plot(df_test_0[pos_key], df_test_0[y_key[j]], marker, color=color[j])

    plt.tight_layout()
    plt.show()
